# -*- coding: utf-8 -*-
"""Module for displaying the server profile in (colored) text format
"""
# import basic stuff
import sys
import enum
from dataclasses import dataclass
from typing import List, Callable, Dict, Optional, Any

# import own stuff
import tlsmate.pdu as pdu
import tlsmate.plugin as plg
import tlsmate.server_profile as server_profile
import tlsmate.tls as tls
import tlsmate.utils as utils
import tlsmate.version as vers

# import other stuff
import colorama  # type: ignore


class Color(tls.ExtendedEnum):
    GREEN = enum.auto()
    RED = enum.auto()
    BLUE = enum.auto()
    WHITE = enum.auto()
    BLACK = enum.auto()
    MAGENTA = enum.auto()
    CYAN = enum.auto()
    YELLOW = enum.auto()


class FontStyle:
    html = False

    def __init__(self, color=None, bold=False):
        self.color = color
        self.bold = bold

    def decorate(self, txt, with_orig_len=False):
        orig_len = len(str(txt))
        if self.html:
            if self.bold:
                txt = f"<b>{txt}</b>"

            if self.color:
                txt = f"<font color={self.color.name.lower()}>{txt}</font>"

        else:
            if self.color:
                txt = f"{getattr(colorama.Fore, self.color.name)}{txt}"

            if self.bold:
                txt = f"{colorama.Style.BRIGHT}{txt}"

            if self.color or self.bold:
                txt = f"{txt}{colorama.Style.RESET_ALL}"

        if with_orig_len:
            return txt, orig_len

        else:
            return txt


@dataclass
class Style:
    GOOD: FontStyle = FontStyle(color=Color.GREEN)
    NEUTRAL: FontStyle = FontStyle()
    SOSO: FontStyle = FontStyle(color=Color.YELLOW, bold=True)
    BAD: FontStyle = FontStyle(color=Color.RED)
    HEADLINE: FontStyle = FontStyle(color=Color.MAGENTA, bold=True)
    BOLD: FontStyle = FontStyle(bold=True)
    ERROR: FontStyle = FontStyle(color=Color.RED, bold=True)


def merge_styles(styles: List[Style]) -> FontStyle:
    """Returns the "worst" style from a given list

    Arguments:
        styles: the list of style strings
    """
    for style in [Style.ERROR, Style.BAD, Style.SOSO, Style.NEUTRAL, Style.GOOD]:
        if style in styles:
            return style

    raise ValueError("cannot merge styles")


def get_dict_value(
    profile: Dict[str, Any], *keys: str, default: Optional[Any] = None
) -> Any:
    """Save function to get an item from a nested dict

    Arguments:
        profile: the dict to start with
        *keys: a list the keys to descent to the wanted item of the profile
        default: the value to return if a key does not exist. Defaults to None.

    Returns:
        the item
    """

    if not profile:
        return default

    for key in keys:
        if key in profile:
            profile = profile[key]

        else:
            return default

    return profile


def get_style(profile: Dict[str, Any], *keys: str) -> FontStyle:
    """Get the style string from a nested dict, descending according to the keys

    The leaf must be in ["good", "neutral", "soso", "bad", "headline", "bold", "reset",
    "error"] (case insensitive). These style names are translated to the corresponding
    ANSI escape code.

    Arguments:
        profile: the nested dict
        *keys: the keys applied in sequence to descent to the style

    Returns:
        ANSI escape code string. If anything fails, Style.ERROR is returned.
    """
    return getattr(
        Style, get_dict_value(profile, *keys, default="error").upper(), Style.ERROR
    )


def get_style_applied(
    txt: str, profile: Dict[str, Any], *keys: str, **kwargs: Any
) -> str:
    """Descent into a nested dict and apply the found style to the given text.

    Arguments:
        txt (str): the text to decorate
        profile (dict): the nested dict
        *keys (str): the keys to descent into the nested dict

    Returns:
        str: the given text, decorated with the found ANSI escape code.
    """

    return get_style(profile, *keys).decorate(txt, **kwargs)


def get_styled_text(data: Dict[str, Any], *path: str, **kwargs: Any) -> str:
    """Comfortable way to use common structure to apply a style to a text.

    The item determined by data and path must be a dict with the following structure:
    {
        "txt": str,
        "style": str
    }
    The text given by the "txt" item will be decorated with the ANSI escape code
    which related to the "style" item.

    Arguments:
        data: the nested dict to use
        *path: the keys to descent into the nested dict

    Returns:
        the "txt" string decorated with the "style" style.
    """
    prof = get_dict_value(data, *path)
    if not prof:
        return Style.ERROR.decorate("???", **kwargs)

    return get_style_applied(get_dict_value(prof, "txt"), prof, "style", **kwargs)


def get_cert_ext(
    cert: server_profile.SPCertificate, name: str
) -> Optional[server_profile.SPCertExtension]:
    """Extract the given extension from a certificate.

    Arguments:
        cert: the certificate object
        name: the name of the extension

    Returns:
        the extension or None if not found
    """
    if not hasattr(cert, "extensions"):
        return None

    for ext in cert.extensions:
        if ext.name == name:
            return ext

    return None


class TextProfileWorker(plg.Worker):
    """Worker class which serializes a server profile.
    """

    name = "text_profile_dumper"
    descr = "dump the scan results"
    prio = 1002

    _callbacks: List[Callable[["TextProfileWorker"], None]] = []

    @classmethod
    def augment_output(cls, callback: Callable) -> Callable:
        """Decorator which can be used to register additional callbacks.

        Arguments:
            callback (callable): the callback to register. After the TextProfileWorker
            is finished with its output, the registered callbacks will be called with
            the TextProfileWorker object as its only argument. No return value is
            expected from the callback.
        """
        cls._callbacks.append(callback)
        return callback

    def _parse_style(self):
        if not self._style:
            return

        if "style" in self._style:
            for style in [
                "good",
                "neutral",
                "soso",
                "bad",
                "headline",
                "bold",
                "error",
            ]:
                style_def = self._style["style"].get(style)
                if style_def is not None:
                    color = None
                    fg = style_def.get("fg")
                    if fg in [c.name.lower() for c in Color.all()]:
                        color = Color.str2enum(fg.upper())

                    bold = style_def.get("style") == "bright"
                    setattr(Style, style.upper(), FontStyle(color=color, bold=bold))

    def _read_style(self):
        self._style = utils.deserialize_data(self.style_file)
        self._parse_style()

    def style_for_cipher_suite(self, cs):
        det = utils.get_cipher_suite_details(cs)
        key_style = get_style(self._style, "key_exchange", det.key_algo.name)
        cipher_style = get_style(self._style, "symmetric_ciphers", det.cipher.name)
        mac_style = get_style(self._style, "macs", det.mac.name)
        return merge_styles([key_style, cipher_style, mac_style])

    def _style_for_assym_key_size(self, bits, style_entry):
        key_sizes = get_dict_value(self._style, style_entry)
        if not key_sizes:
            return Style.NEUTRAL

        for prof in key_sizes:
            if bits >= prof["size"]:
                break

        return get_style(prof, "style")

    def style_for_rsa_dh_key_size(self, bits):
        return self._style_for_assym_key_size(bits, "assymetric_key_sizes")

    def style_for_ec_key_size(self, bits):
        return self._style_for_assym_key_size(bits, "assymetric_ec_key_sizes")

    def _print_tlsmate(self):
        print(Style.HEADLINE.decorate("A TLS configuration scanner (and more)"))
        print()
        table = utils.Table(indent=2, sep="  ")
        table.row("tlsmate version", vers.__version__)
        table.row("repository", "https://gitlab.com/guballa/tlsmate")
        table.dump()
        print(
            "  Please file bug reports at https://gitlab.com/guballa/tlsmate/-/issues"
        )
        print()

    def _print_scan_info(self):
        if not hasattr(self.server_profile, "scan_info"):
            return

        scan_info = self.server_profile.scan_info
        self._start_date = scan_info.start_date
        print(Style.HEADLINE.decorate("Basic scan information"))
        print()
        print(f"  command: {scan_info.command}")
        table = utils.Table(indent=2, sep="  ")
        table.row("tlsmate version", f"{scan_info.version} (producing the scan)")
        table.row("scan start timestamp", str(scan_info.start_date))
        table.row("scan duration", f"{scan_info.run_time} seconds")
        table.row("applied style", self.style_file)
        table.row(
            "style description",
            get_dict_value(self._style, "description", "short", default="-"),
        )
        table.dump()
        print()

    def _print_host(self):
        if not hasattr(self.server_profile, "server"):
            return

        host_info = self.server_profile.server
        print(Style.HEADLINE.decorate("Scanned host"))
        print()
        table = utils.Table(indent=2, sep="  ")
        name_resolution = hasattr(host_info, "name_resolution")
        if name_resolution:
            host = host_info.name_resolution.domain_name

        else:
            host = host_info.ip

        table.row("host", host)
        table.row("port", str(host_info.port))
        table.row("SNI", host_info.sni)
        if name_resolution:
            if hasattr(host_info.name_resolution, "ipv4_addresses"):
                addresses = ", ".join(host_info.name_resolution.ipv4_addresses)
                table.row("IPv4 addresses", addresses)

            if hasattr(host_info.name_resolution, "ipv6_addresses"):
                addresses = ", ".join(host_info.name_resolution.ipv6_addresses)
                table.row("IPv6 addresses", addresses)

        proxy = getattr(host_info, "proxy", None)
        if proxy:
            table.row("HTTP-proxy", proxy)

        table.dump()
        print()

    def _print_server_malfunctions(self):
        malfunctions = getattr(self.server_profile, "server_malfunctions", None)
        if not malfunctions:
            return

        print(Style.HEADLINE.decorate("Severe server implementation flaws"))
        print()
        for malfunction in malfunctions:
            add_info = []
            if hasattr(malfunction, "message"):
                add_info.append(f"message: {malfunction.message.name}")

            if hasattr(malfunction, "extension"):
                add_info.append(f"extension: {malfunction.extension.name}")

            txt = "  - " + get_dict_value(
                self._style,
                "server_malfunction",
                malfunction.issue.name,
                default=malfunction.issue.description,
            )
            if add_info:
                txt += f" ({'; '.join(add_info)})"

            print(Style.BAD.decorate(txt))

        print()

    def _print_versions(self):
        if not hasattr(self.server_profile, "versions"):
            return

        print(Style.HEADLINE.decorate("TLS protocol versions"))
        print()
        table = utils.Table(indent=2, sep="  ")
        for version_prof in self.server_profile.versions:
            table.row(
                version_prof.version.name,
                get_styled_text(
                    self._style,
                    "version",
                    version_prof.version.name,
                    "supported",
                    version_prof.support.name,
                ),
            )

        table.dump()
        print()

    def _print_cipher_suites(self):
        if not hasattr(self.server_profile, "versions"):
            return

        cipher_hash = {}
        print(Style.HEADLINE.decorate("Cipher suites"))

        for version in self._prof_versions:
            version_prof = self.server_profile.get_version_profile(version)
            if version is tls.Version.SSL20:
                cipher_list = version_prof.cipher_kinds
                pref_txt = ""
                style_txt = (
                    ""
                    if cipher_list
                    else Style.BAD.decorate("no cipher kinds provided by server")
                )
                chacha_txt = ""
            else:
                cipher_list = version_prof.ciphers.cipher_suites
                order = version_prof.ciphers.server_preference

                struct = get_dict_value(
                    self._style, "version", version.name, "cipher_order", order.name
                )
                pref_txt = get_dict_value(struct, "txt", default="???")
                style_txt = get_style_applied(pref_txt, struct, "style")

                chacha_pref = getattr(
                    version_prof.ciphers, "chacha_poly_preference", None
                )
                if chacha_pref:
                    struct = get_dict_value(
                        self._style,
                        "version",
                        version.name,
                        "chacha_preference",
                        chacha_pref.name,
                    )
                    chacha_txt = get_style_applied(
                        get_dict_value(struct, "txt", default="???"), struct, "style"
                    )

                else:
                    chacha_txt = ""

            hashed = hash((style_txt, chacha_txt, tuple(cipher_list)))
            if hashed in cipher_hash:
                cipher_hash[hashed]["versions"].append(str(version))

            else:
                cipher_hash[hashed] = {
                    "versions": [str(version)],
                    "table": utils.Table(indent=4, sep="  "),
                    "preference": style_txt,
                    "chacha_preference": chacha_txt,
                }
                all_good = bool(cipher_list)
                for cs in cipher_list:
                    if version is tls.Version.SSL20:
                        cipher_hash[hashed]["table"].row(
                            f"0x{cs.value:06x}", Style.BAD.decorate(cs)
                        )

                    else:
                        style = self.style_for_cipher_suite(cs)
                        if style is not Style.GOOD:
                            all_good = False

                        cipher_hash[hashed]["table"].row(
                            f"0x{cs.value:04x}", style.decorate(cs.name)
                        )

                if all_good:
                    cipher_hash[hashed]["preference"] = Style.NEUTRAL.decorate(pref_txt)

        for values in cipher_hash.values():
            versions = Style.BOLD.decorate(", ".join(values["versions"]))
            print(f"\n  {versions}:")
            print(f'    {values["preference"]}')
            if values["chacha_preference"] != "":
                print(f'    {values["chacha_preference"]}')

            values["table"].dump()

        print()

    def _print_supported_groups(self):
        if not hasattr(self.server_profile, "versions"):
            return

        prof_grps = get_dict_value(self._style, "supported_groups", "groups")
        group_hash = {}
        for version in self._prof_versions:
            prof_version = get_dict_value(self._style, "supported_groups", version.name)
            version_prof = self.server_profile.get_version_profile(version)
            group_prof = getattr(version_prof, "supported_groups", None)
            if group_prof is None:
                continue

            if not hasattr(group_prof, "groups"):
                continue

            supported = getattr(group_prof, "extension_supported", None)
            if supported is None:
                supp_txt = None

            else:
                prof = get_dict_value(prof_version, "support", supported.name)
                supp_txt = get_dict_value(prof, "txt", default="???")
                if supp_txt:
                    supp_txt = get_style_applied(supp_txt, prof, "style")

            preference = getattr(group_prof, "server_preference", None)
            if preference is None:
                pref_txt = None

            else:
                prof = get_dict_value(prof_version, "preference", preference.name)
                pref_txt = get_dict_value(prof, "txt", default="???")
                if pref_txt:
                    if prof_grps and all(
                        [prof_grps[grp.name] == "good" for grp in group_prof.groups]
                    ):
                        pref_style = Style.NEUTRAL

                    else:
                        pref_style = get_style(prof, "style")

                    pref_txt = pref_style.decorate(pref_txt)

            advertised = getattr(group_prof, "groups_advertised", None)
            if advertised is None:
                ad_txt = None

            else:
                prof = get_dict_value(prof_version, "advertised", advertised.name)
                ad_txt = get_dict_value(prof, "txt", default="???")
                if ad_txt:
                    ad_txt = get_style_applied(ad_txt, prof, "style")

            combined = (supp_txt, pref_txt, ad_txt, tuple(group_prof.groups))
            hashed = hash(combined)
            if hashed in group_hash:
                group_hash[hashed]["versions"].append(str(version))

            else:
                group_hash[hashed] = {"versions": [str(version)], "combined": combined}

        if not group_hash:
            return

        print(Style.HEADLINE.decorate("Supported groups"))
        for group in group_hash.values():
            versions = ", ".join(group["versions"])
            print(f"\n  {Style.BOLD.decorate(versions)}:")
            supp_txt, pref_txt, ad_txt, groups = group["combined"]

            if supp_txt:
                print(f"    {supp_txt}")

            if pref_txt:
                print(f"    {pref_txt}")

            if ad_txt:
                print(f"    {ad_txt}")

            print("    supported groups:")
            table = utils.Table(indent=6, sep="  ")
            for grp in groups:
                table.row(
                    f"0x{grp.value:02x}",
                    get_style_applied(grp.name, prof_grps, grp.name),
                )

            table.dump()

        print()

    def _print_sig_algos(self):
        if not hasattr(self.server_profile, "versions"):
            return

        algo_hash = {}
        for version in self._prof_versions:
            version_prof = self.server_profile.get_version_profile(version)
            if not hasattr(version_prof, "signature_algorithms"):
                continue

            algo_prof = version_prof.signature_algorithms
            hashed = hash(tuple(algo_prof.algorithms))
            if hashed in algo_hash:
                algo_hash[hashed]["versions"].append(str(version))

            else:
                algo_hash[hashed] = {
                    "versions": [str(version)],
                    "algos": algo_prof.algorithms,
                }

        if not algo_hash:
            return

        print(Style.HEADLINE.decorate("Signature algorithms"))
        for algo in algo_hash.values():
            versions = ", ".join(algo["versions"])
            print(f"\n  {Style.BOLD.decorate(versions)}:")

            if not algo["algos"]:
                print("    no signature algorithms supported")

            else:
                print("    signature algorithms:")
                table = utils.Table(indent=6, sep="  ")
                if algo["algos"]:
                    for alg in algo["algos"]:
                        table.row(
                            f"0x{alg.value:04x}",
                            get_style_applied(
                                alg.name, self._style, "signature_schemes", alg.name
                            ),
                        )

                table.dump()

        print()

    def _print_dh_groups(self):
        if not hasattr(self.server_profile, "versions"):
            return

        dh_groups = {}
        for version in self._prof_versions:
            version_prof = self.server_profile.get_version_profile(version)
            dh_prof = getattr(version_prof, "dh_group", None)
            if dh_prof is None:
                continue

            name = getattr(dh_prof, "name", None)
            combined = (name, dh_prof.size)
            hashed = hash(combined)
            if hashed in dh_groups:
                dh_groups[hashed]["versions"].append(str(version))

            else:
                dh_groups[hashed] = {
                    "versions": [str(version)],
                    "combined": combined,
                }

        if dh_groups:
            print(Style.HEADLINE.decorate("DH groups (finite field)"))
            for values in dh_groups.values():
                versions = ", ".join(values["versions"])
                print(f"\n  {Style.BOLD.decorate(versions)}:")
                name, size = values["combined"]
                if name is None:
                    name = "unknown group"
                txt = f"{name} ({size} bits)"
                style = self.style_for_rsa_dh_key_size(size)
                print(f"    {style.decorate(txt)}")

            print()

    def _print_common_features(self, feat_prof):

        if not any(
            hasattr(feat_prof, prop)
            for prop in [
                "ocsp_stapling",
                "ocsp_multi_stapling",
                "heartbeat",
                "downgrade_attack_prevention",
            ]
        ):
            return

        print(f'  {Style.BOLD.decorate("Common features")}')
        table = utils.Table(indent=4, sep="  ")

        ocsp_state = getattr(feat_prof, "ocsp_stapling", None)
        if ocsp_state is not None:
            table.row(
                "OCSP stapling (status_request)",
                get_styled_text(self._style, "ocsp_stapling", ocsp_state.name),
            )

        ocsp_state = getattr(feat_prof, "ocsp_multi_stapling", None)
        if ocsp_state is not None:
            table.row(
                "OCSP multi stapling (status_request_v2)",
                get_styled_text(self._style, "ocsp_multi_stapling", ocsp_state.name),
            )

        hb_state = getattr(feat_prof, "heartbeat", None)
        if hb_state is not None:
            table.row(
                "Heartbeat", get_styled_text(self._style, "heartbeat", hb_state.name)
            )

        fallback = getattr(feat_prof, "downgrade_attack_prevention", None)
        if fallback is not None:
            table.row(
                "Downgrade attack prevention",
                get_styled_text(self._style, "fallback", fallback.name),
            )

        table.dump()
        print()

    def _print_features_tls12(self, feat_prof):
        if not any(
            hasattr(feat_prof, prop)
            for prop in [
                "compression",
                "scsv_renegotiation",
                "encrypt_then_mac",
                "extended_master_secret",
                "insecure_renegotiation",
                "secure_renegotation",
                "session_id",
                "session_ticket",
            ]
        ):
            return

        print(f'  {Style.BOLD.decorate("Features for TLS1.2 and below")}')
        table = utils.Table(indent=4, sep="  ")
        if hasattr(feat_prof, "compression"):
            if (len(feat_prof.compression) == 1) and feat_prof.compression[
                0
            ] is tls.CompressionMethod.NULL:
                compr = tls.ScanState.FALSE

            else:
                compr = tls.ScanState.TRUE

            table.row(
                "compression", get_styled_text(self._style, "compression", compr.name)
            )

        etm = getattr(feat_prof, "encrypt_then_mac", None)
        if etm is not None:
            table.row(
                "encrypt-then-mac",
                get_styled_text(self._style, "encrypt_then_mac", etm.name),
            )

        ems = getattr(feat_prof, "extended_master_secret", None)
        if ems is not None:
            table.row(
                "extended master secret",
                get_styled_text(self._style, "extended_master_secret", ems.name),
            )

        insec_reneg = getattr(feat_prof, "insecure_renegotiation", None)
        if insec_reneg is not None:
            table.row(
                "insecure renegotiation",
                get_styled_text(
                    self._style, "insecure_renegotiation", insec_reneg.name
                ),
            )

        sec_reneg = getattr(feat_prof, "secure_renegotation", None)
        if sec_reneg is not None:
            table.row(
                "secure renegotiation (extension)",
                get_styled_text(self._style, "secure_renegotiation", sec_reneg.name),
            )

        scsv = getattr(feat_prof, "scsv_renegotiation", None)
        if scsv is not None:
            table.row(
                "secure renegotiation (SCSV)",
                get_styled_text(self._style, "scsv_renegotiation", scsv.name),
            )

        session_id = getattr(feat_prof, "session_id", None)
        if session_id is not None:
            table.row(
                "resumption with session_id",
                get_styled_text(self._style, "session_id", session_id.name),
            )

        session_ticket = getattr(feat_prof, "session_ticket", None)
        if session_ticket is not None:
            txt = get_styled_text(self._style, "session_ticket", session_ticket.name)
            life_time = getattr(feat_prof, "session_ticket_lifetime", None)
            if life_time is None:
                add_txt = ""

            else:
                add_txt = f", life time: {feat_prof.session_ticket_lifetime} seconds"

            table.row("resumption with session ticket", f"{txt}{add_txt}")

        table.dump()
        print()

    def _print_features_tls13(self, feat_prof):
        if not any(
            hasattr(feat_prof, prop) for prop in ["resumption_psk", "early_data"]
        ):
            return

        print(f'  {Style.BOLD.decorate("Features for TLS1.3")}')

        table = utils.Table(indent=4, sep="  ")
        resumption_psk = getattr(feat_prof, "resumption_psk", None)
        if resumption_psk is not None:
            txt = get_styled_text(self._style, "resumption_psk", resumption_psk.name)
            life_time = getattr(feat_prof, "psk_lifetime", None)
            if life_time is None:
                add_txt = ""

            else:
                add_txt = f", life time: {feat_prof.psk_lifetime} seconds"

            table.row("resumption with PSK", f"{txt}{add_txt}")

        early_data = getattr(feat_prof, "early_data", None)
        if early_data is not None:
            table.row(
                "early data (0-RTT)",
                get_styled_text(self._style, "early_data", early_data.name),
            )

        table.dump()
        print()

    def _print_grease(self, grease_prof):
        grease = (
            ("version_tolerance", "version"),
            ("cipher_suite_tolerance", "cipher_suite"),
            ("extension_tolerance", "extension"),
            ("group_tolerance", "group"),
            ("sig_algo_tolerance", "sig_algo"),
            ("psk_mode_tolerance", "psk_mode"),
        )

        caption = Style.BOLD.decorate(
            "Server tolerance to unknown values (GREASE, RFC8701)"
        )
        print(f"  {caption}")
        table = utils.Table(indent=4, sep="  ")

        for prof_prop, style_prop in grease:
            val = getattr(grease_prof, prof_prop, tls.ScanState.UNDETERMINED)
            table.row(
                get_dict_value(
                    self._style, "grease", style_prop, "descr", default="???"
                ),
                get_styled_text(self._style, "grease", style_prop, val.name),
            )

        table.dump()
        print()

    def _print_ephemeral_key_reuse(self):
        if not hasattr(self.server_profile.features, "ephemeral_key_reuse"):
            return

        ekr = self.server_profile.features.ephemeral_key_reuse
        print(Style.BOLD.decorate("  Ephemeral key reuse"))
        table = utils.Table(indent=4, sep="  ")
        table.row(
            "DHE key reuse (TLS1.2 or below)",
            get_styled_text(
                self._style, "ephemeral_key_reuse", ekr.tls12_dhe_reuse.name
            ),
        )
        table.row(
            "ECDHE key reuse (TLS1.2 or below)",
            get_styled_text(
                self._style, "ephemeral_key_reuse", ekr.tls12_ecdhe_reuse.name
            ),
        )
        table.row(
            "DHE key reuse (TLS1.3)",
            get_styled_text(
                self._style, "ephemeral_key_reuse", ekr.tls13_dhe_reuse.name
            ),
        )
        table.row(
            "ECDHE key reuse (TLS1.3)",
            get_styled_text(
                self._style, "ephemeral_key_reuse", ekr.tls13_ecdhe_reuse.name
            ),
        )
        table.dump()
        print()

    def _print_features(self):
        feat_prof = getattr(self.server_profile, "features", None)
        if not feat_prof:
            return

        print(Style.HEADLINE.decorate("Features"))
        print()

        self._print_common_features(feat_prof)

        versions = self.server_profile.get_versions()
        wanted = {
            tls.Version.SSL30,
            tls.Version.TLS10,
            tls.Version.TLS11,
            tls.Version.TLS12,
        }
        if wanted.intersection(set(versions)):
            self._print_features_tls12(feat_prof)

        if tls.Version.TLS13 in versions:
            self._print_features_tls13(feat_prof)

        if hasattr(feat_prof, "grease"):
            self._print_grease(feat_prof.grease)

        self._print_ephemeral_key_reuse()

    def _print_cert(self, cert, idx):
        items = [str(getattr(cert, "version", ""))]
        self_signed = getattr(cert, "self_signed", None)
        if self_signed is tls.ScanState.TRUE:
            items.append("self-signed")

        from_trust_store = getattr(cert, "from_trust_store", tls.ScanState.FALSE)
        if from_trust_store is tls.ScanState.TRUE:
            items.append("certificate taken from trust store")

        print(f'  Certificate #{idx}: {", ".join(items)}')
        table = utils.Table(indent=4, sep="  ")

        issues = getattr(cert, "issues", None)
        if issues:
            issue_txt = []
            style = get_style(self._style, "certificate", "issues")
            for issue in issues:
                folded_lines = utils.fold_string(issue, max_length=100)
                issue_txt.append("- " + folded_lines.pop(0))
                issue_txt.extend(["  " + item for item in folded_lines])

            table.row("Issues", style.decorate(issue_txt[0]))
            for line in issue_txt[1:]:
                table.row("", style.decorate(line))

        table.row("Serial number", f"{cert.serial_number_int} (integer)")
        table.row("", f"{pdu.string(cert.serial_number_bytes)} (hex)")
        lines = utils.fold_string(cert.subject, max_length=100, sep=",")
        table.row("Subject", lines.pop(0))
        for line in lines:
            table.row("", line)

        san_ext = get_cert_ext(cert, "SubjectAlternativeName")
        if san_ext is not None:
            sans = " ".join([str(name) for name in san_ext.subj_alt_names])
            lines = utils.fold_string(sans, max_length=80)
            table.row("SubjectAltName (SAN)", lines.pop(0))
            for line in lines:
                table.row("", line)

        if hasattr(cert, "subject_matches"):
            table.row(
                "URI matches",
                get_styled_text(
                    self._style,
                    "certificate",
                    "subject_matches",
                    cert.subject_matches.name,
                ),
            )

        ev = getattr(cert, "extended_validation", tls.ScanState.NA)
        if ev is not tls.ScanState.NA:
            table.row(
                "Extended validation",
                get_styled_text(
                    self._style, "certificate", "extended_validation", ev.name
                ),
            )

        lines = utils.fold_string(cert.issuer, max_length=100, sep=",")
        table.row("Issuer", lines.pop(0))
        for line in lines:
            table.row("", line)

        sig_algo = getattr(cert, "signature_algorithm", None)
        if sig_algo is not None:
            if self_signed is tls.ScanState.TRUE:
                txt = sig_algo.name

            else:
                txt = get_style_applied(
                    sig_algo.name, self._style, "signature_schemes", sig_algo.name
                )

            table.row("Signature algorithm", txt)

        pub_key = getattr(cert, "public_key", None)
        if pub_key is not None:
            key_size = pub_key.key_size
            if pub_key.key_type in [
                tls.SignatureAlgorithm.RSA,
                tls.SignatureAlgorithm.DSA,
            ]:
                style = self.style_for_rsa_dh_key_size(key_size)

            else:
                style = self.style_for_ec_key_size(key_size)

            table.row(
                "Public key",
                f'{pub_key.key_type}, {style.decorate(f"{key_size} bits")}',
            )

        key_usage_ext = get_cert_ext(cert, "KeyUsage")
        if key_usage_ext is not None:
            usage_txt = ", ".join([str(usage) for usage in key_usage_ext.key_usage])
            table.row("Key usage", usage_txt)

        ext_key_usage_ext = get_cert_ext(cert, "ExtendedKeyUsage")
        if ext_key_usage_ext is not None:
            usage_txt = ", ".join(
                [str(usage) for usage in ext_key_usage_ext.extended_key_usage]
            )
            table.row("Extended key usage", usage_txt)

        if hasattr(cert, "not_valid_before"):
            valid = tls.ScanState.TRUE
            valid_from = tls.ScanState.TRUE
            if cert.not_valid_before > self._start_date:
                valid_from = tls.ScanState.FALSE
                valid = tls.ScanState.FALSE

            valid_to = tls.ScanState.TRUE
            if cert.not_valid_after < self._start_date:
                valid_to = tls.ScanState.FALSE
                valid = tls.ScanState.FALSE

            from_txt = get_style_applied(
                cert.not_valid_before,
                self._style,
                "certificate",
                "validity",
                valid_from.name,
                "style",
            )
            to_txt = get_style_applied(
                cert.not_valid_after,
                self._style,
                "certificate",
                "validity",
                valid_to.name,
                "style",
            )
            valid_txt = get_styled_text(
                self._style, "certificate", "validity", valid.name
            )
            table.row(
                "Validity period",
                (
                    f"{from_txt} - {to_txt} ({cert.validity_period_days} days), "
                    f"{valid_txt}"
                ),
            )

        crl_distr = get_cert_ext(cert, "CRLDistributionPoints")
        if crl_distr is not None:
            crls = []
            for distr_schema in crl_distr.distribution_points:
                for gen_name in distr_schema.full_name:
                    if hasattr(gen_name, "uri"):
                        crls.append(gen_name.uri)

            table.row("CRLs", crls.pop(0))
            for crl in crls:
                table.row("", crl)

        crl_status = getattr(cert, "crl_revocation_status", None)
        if crl_status is not None:
            table.row(
                "CRL revocation status",
                get_styled_text(
                    self._style, "certificate", "crl_status", crl_status.name
                ),
            )

        ocsp_status = getattr(cert, "ocsp_revocation_status", None)
        if ocsp_status is not None:
            table.row(
                "OCSP revocation status",
                get_styled_text(
                    self._style, "certificate", "ocsp_status", ocsp_status.name
                ),
            )

            text = []
            must_staple = tls.ScanState.FALSE
            if cert.ocsp_must_staple is tls.ScanState.TRUE:
                text.append("must staple")
                must_staple = tls.ScanState.TRUE

            if cert.ocsp_must_staple_multi is tls.ScanState.TRUE:
                text.append("must multi-staple")
                must_staple = tls.ScanState.TRUE

            txt = get_dict_value(
                self._style,
                "certificate",
                "must_staple",
                must_staple.name,
                "txt",
                default="???",
            )
            if text:
                txt += f" ({', '.join(text)})"

            table.row(
                "OCSP must staple",
                get_style_applied(
                    txt,
                    self._style,
                    "certificate",
                    "must_staple",
                    must_staple.name,
                    "style",
                ),
            )

        if hasattr(cert, "fingerprint_sha1"):
            table.row("Fingerprint SHA1", pdu.string(cert.fingerprint_sha1))

        if hasattr(cert, "fingerprint_sha256"):
            table.row("Fingerprint SHA256", pdu.string(cert.fingerprint_sha256))

        table.dump()
        print()

    def _print_certificates(self):
        cert_chains = getattr(self.server_profile, "cert_chains", None)
        if not cert_chains:
            return

        print(Style.HEADLINE.decorate("Certificate chains"))
        for cert_chain in cert_chains:
            print()
            if hasattr(cert_chain, "successful_validation"):
                valid_txt = get_styled_text(
                    self._style,
                    "cert_chain",
                    "validation",
                    cert_chain.successful_validation.name,
                )

            else:
                valid_txt = ""

            head_line = Style.BOLD.decorate(f"Certificate chain #{cert_chain.id}:")
            print(f"  {head_line} {valid_txt}")
            if hasattr(cert_chain, "issues"):
                style = get_style(self._style, "cert_chain", "issues")
                print("    Issues:")
                for issue in cert_chain.issues:
                    lines = utils.fold_string(issue, max_length=100)
                    txt = "    - " + "\n      ".join(lines)
                    print(style.decorate(txt))

            if hasattr(cert_chain, "root_cert_transmitted"):
                txt = get_styled_text(
                    self._style,
                    "cert_chain",
                    "root_cert_transmitted",
                    cert_chain.root_cert_transmitted.name,
                )
                print("    ", txt)

            for idx, cert in enumerate(cert_chain.cert_chain, start=1):
                self._print_cert(cert, idx)

            root_cert = getattr(cert_chain, "root_certificate", None)
            if root_cert:
                self._print_cert(root_cert, len(cert_chain.cert_chain) + 1)

    def _print_vulnerabilities(self):
        vuln_prof = getattr(self.server_profile, "vulnerabilities", None)
        if not vuln_prof:
            return

        table = utils.Table(indent=2, sep="  ")
        print(Style.HEADLINE.decorate("Vulnerabilities"))
        print()

        beast = getattr(vuln_prof, "beast", None)
        if beast is not None:
            table.row(
                "BEAST (CVE-2011-3389)",
                get_styled_text(self._style, "vulnerabilities", "beast", beast.name),
            )

        ccs = getattr(vuln_prof, "ccs_injection", None)
        if ccs is not None:
            table.row(
                "CCS injection (CVE-2014-0224)",
                get_styled_text(
                    self._style, "vulnerabilities", "ccs_injection", ccs.name
                ),
            )

        crime = getattr(vuln_prof, "crime", None)
        if crime is not None:
            table.row(
                "CRIME (CVE-2012-4929)",
                get_styled_text(self._style, "vulnerabilities", "crime", crime.name),
            )

        freak = getattr(vuln_prof, "freak", None)
        if freak is not None:
            table.row(
                "FREAK (CVE-2015-0204)",
                get_styled_text(self._style, "vulnerabilities", "freak", freak.name),
            )

        hb = getattr(vuln_prof, "heartbleed", None)
        if hb is not None:
            table.row(
                "Heartbleed (CVE-2014-0160)",
                get_styled_text(self._style, "vulnerabilities", "heartbleed", hb.name),
            )

        logjam = getattr(vuln_prof, "logjam", None)
        if logjam is not None:
            table.row(
                "Logjam (CVE-2015-0204)",
                get_styled_text(self._style, "vulnerabilities", "logjam", logjam.name),
            )

        robot = getattr(vuln_prof, "robot", None)
        if robot is not None:
            table.row(
                "ROBOT (CVE-2017-13099, ...)",
                get_styled_text(self._style, "vulnerabilities", "robot", robot.name),
            )

        sweet_32 = getattr(vuln_prof, "sweet_32", None)
        if sweet_32 is not None:
            table.row(
                "Sweet32 (CVE-2016-2183, CVE-2016-6329)",
                get_styled_text(
                    self._style, "vulnerabilities", "sweet_32", sweet_32.name
                ),
            )

        poodle = getattr(vuln_prof, "poodle", None)
        if poodle is not None:
            table.row(
                "POODLE (CVE-2014-3566)",
                get_styled_text(self._style, "vulnerabilities", "poodle", poodle.name),
            )

        tls_poodle = getattr(vuln_prof, "tls_poodle", None)
        if tls_poodle is not None:
            table.row(
                "TLS POODLE",
                get_styled_text(
                    self._style, "vulnerabilities", "tls_poodle", tls_poodle.name
                ),
            )

        lucky_minus_20 = getattr(vuln_prof, "lucky_minus_20", None)
        if lucky_minus_20 is not None:
            table.row(
                "Lucky-Minus-20 (CVE-2016-2107)",
                get_styled_text(
                    self._style,
                    "vulnerabilities",
                    "lucky_minus_20",
                    lucky_minus_20.name,
                ),
            )

        # cbc padding oracle shall be the last vulnerability in the list
        cbc_padding_oracle = getattr(vuln_prof, "cbc_padding_oracle", None)
        if cbc_padding_oracle:
            style_cbc = get_dict_value(
                self._style, "vulnerabilities", "cbc_padding_oracle"
            )
            vulnerable = getattr(cbc_padding_oracle, "vulnerable", None)
            txt = get_dict_value(
                style_cbc, "vulnerable", vulnerable.name, "txt", default="???"
            )
            oracles = getattr(cbc_padding_oracle, "oracles", None)
            if oracles:
                txt += f", number of oracles: {len(oracles)}"

            table.row(
                "CBC padding oracle",
                get_style_applied(
                    txt, style_cbc, "vulnerable", vulnerable.name, "style"
                ),
            )
            accuracy = getattr(cbc_padding_oracle, "accuracy", None)
            table.row(
                "  scan accuracy", get_styled_text(style_cbc, "accuracy", accuracy.name)
            )

        table.dump()
        if cbc_padding_oracle and oracles:
            style_oracle = get_dict_value(style_cbc, "oracle")
            for oracle in oracles:
                print("\n    oracle properties")
                table = utils.Table(indent=6, sep="  ")
                strong = getattr(oracle, "strong", None)
                table.row(
                    "strength", get_styled_text(style_oracle, "strong", strong.name),
                )
                observable = getattr(oracle, "observable", None)
                table.row(
                    "observable",
                    get_styled_text(style_oracle, "observable", observable.name),
                )
                oracle_types = getattr(oracle, "types", None)
                if oracle_types:
                    str_types = []
                    for oracle_type in oracle_types:
                        str_types.append(
                            get_dict_value(
                                style_oracle, "type", oracle_type.name, default="???"
                            )
                        )

                    table.row("oracle type(s)", str_types.pop(0))
                    for line in str_types:
                        table.row("", line)

                cipher_groups = getattr(oracle, "cipher_group", None)
                if cipher_groups:
                    str_group = []
                    max_len = 0
                    for group in cipher_groups:
                        version = getattr(group, "version", None)
                        version_str = version.name if version else "???"
                        cs = getattr(group, "cipher_suite", None)
                        cs_str = cs.name if cs else "???"
                        protocol = getattr(group, "record_protocol", None)
                        protocol_str = get_dict_value(
                            style_oracle,
                            "cipher_group",
                            "record_protocol",
                            protocol.name,
                        )
                        str_group.append((version_str, cs_str, protocol_str))
                        max_len = max(max_len, len(cs_str))
                        # str_group.append(f"  {version_str} {cs_str}   {protocol_str}")

                    txt = "cipher suite groups"
                    for version, cs, prot in str_group:
                        table.row(txt, f"{version} {cs:{max_len}} {prot}")
                        txt = ""

                table.dump()

        print()

    def run(self):
        self.style_file = self.config.get("style")
        self._read_style()
        if self.config.get("progress"):
            sys.stderr.write("\n")

        if self.config.get("format") == "html":
            FontStyle.html = True
            print("<pre>")

        else:
            colorama.init(strip=not self.config.get("color"))

        self._prof_versions = self.server_profile.get_versions()
        self._print_tlsmate()
        self._print_scan_info()
        self._print_host()
        self._print_server_malfunctions()
        self._print_versions()
        if self._prof_versions:
            self._print_cipher_suites()
            self._print_supported_groups()
            self._print_sig_algos()
            self._print_dh_groups()
            self._print_features()
            self._print_certificates()
            self._print_vulnerabilities()

        for callback in self._callbacks:
            callback(self)

        if self.config.get("format") == "html":
            print("</pre>")



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.robot import ScanRobot
from tlsmate.tlssuite import TlsSuiteTester, TlsLibrary


class TestCaseWeakOracle(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_basic_wolfssl3_12_0"
    recorder_yaml = "recorder_robot_weak_oracle"
    path = pathlib.Path(__file__)
    server_cmd = "utils/start_wolfssl --version {library} --port {server_port}"

    library = TlsLibrary.wolfssl3_12_0

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        tlsmate.client.alert_on_invalid_cert = False
        ScanRobot(tlsmate).run()
        profile = server_profile.make_serializable()
        assert profile["vulnerabilities"]["robot"] == "WEAK_ORACLE"


if __name__ == "__main__":
    TestCaseWeakOracle().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate import msg
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.

    Note: The openssl server must be started with "-early_data" and we cannot use
    this option in combination with "-www".
    """

    name = "EarlyData"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa "
        "-- -cipher ALL -early_data"
    )
    library = TlsLibrary.openssl1_1_1

    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    # version = tls.Version.TLS12

    def run(self, tlsmate, is_replaying=False):
        client = tlsmate.client
        client.init_profile()

        client.profile.versions = [tls.Version.TLS13]
        client.profile.cipher_suites = [tls.CipherSuite.TLS_AES_128_GCM_SHA256]
        client.profile.supported_groups = [
            tls.SupportedGroups.SECP256R1,
            tls.SupportedGroups.SECP384R1,
            tls.SupportedGroups.SECP521R1,
            tls.SupportedGroups.X25519,
            tls.SupportedGroups.X448,
        ]
        client.profile.key_shares = self.supported_groups
        client.profile.signature_algorithms = [
            tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
            tls.SignatureScheme.RSA_PKCS1_SHA256,
            tls.SignatureScheme.RSA_PSS_PSS_SHA256,
            tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
        ]
        client.profile.support_psk = True
        client.profile.psk_key_exchange_modes = [tls.PskKeyExchangeMode.PSK_DHE_KE]
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.handshake()
            conn.wait(msg.NewSessionTicket)
            end_of_tc_reached = True
        assert end_of_tc_reached is True

        client.profile.early_data = b"This is EarlyData (0-RTT)"
        end_of_tc_reached = False
        with client.create_connection() as conn:
            conn.handshake()
            end_of_tc_reached = True
        assert end_of_tc_reached is True
        assert conn.early_data_accepted is True


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module for a worker handling the server profile (de)serialization
"""
# import basic stuff

# import own stuff
import tlsmate.plugin as plg
import tlsmate.utils as utils

# import other stuff


class ReadProfileWorker(plg.Worker):
    """Worker class which deserializes a server profile in Yaml format
    """

    name = "profile_reader"
    descr = "read the server profile"
    prio = 1

    def run(self):
        read_profile = self.config.get("read_profile")
        if read_profile is not None:
            self.server_profile.load(utils.deserialize_data(read_profile))


class DumpProfileWorker(plg.Worker):
    """Worker class which serializes a server profile.
    """

    name = "profile_dumper"
    descr = "dump the server profile"
    prio = 1001

    def run(self):
        utils.serialize_data(
            self.server_profile.make_serializable(),
            file_name=self.config.get("write_profile"),
            replace=True,
            use_json=(self.config.get("format") == "json"),
            indent=4,
        )



# -*- coding: utf-8 -*-
"""Module providing infos about the scanner
"""
# import basic stuff
import sys
import time
import datetime

# import own stuff
import tlsmate.plugin as plg
import tlsmate.resolver as resolver
import tlsmate.server_profile as server_profile
import tlsmate.tls as tls
import tlsmate.version as version

# import other stuff


class ScanStart(plg.Worker):
    """Provide basic infos without actually really scanning against the server.
    """

    name = "scanstart"
    descr = "determine basic settings"
    prio = 0

    def run(self):
        """The entry point for the worker.
        """

        if not hasattr(self.server_profile, "scan_info"):
            self.server_profile.scan_info = server_profile.SPScanInfo()

        scan_info = self.server_profile.scan_info
        start_time = time.time()
        scan_info.command = " ".join(sys.argv)
        scan_info.version = version.__version__
        scan_info.start_timestamp = start_time
        scan_info.start_date = datetime.datetime.fromtimestamp(int(start_time))
        endp = resolver.determine_l4_addr(
            self.config.get("host"), self.config.get("port")
        )
        data = {"port": endp.port}
        if endp.host_type is tls.HostType.HOST:
            name_res_data = {"domain_name": endp.host}
            ips = resolver.resolve_hostname(endp.host, self.config.get("proxy"))
            if ips.ipv4_addresses:
                name_res_data["ipv4_addresses"] = ips.ipv4_addresses

            if ips.ipv6_addresses:
                name_res_data["ipv6_addresses"] = ips.ipv6_addresses

            endp = resolver.get_ip_endpoint(endp)
            data["name_resolution"] = server_profile.SPNameResolution(
                data=name_res_data
            )

        data["ip"] = endp.host
        try:
            data["sni"] = self.client.get_sni()

        except ValueError:
            pass

        proxy = self.config.get("proxy")
        if proxy:
            data["proxy"] = proxy

        self.server_profile.server = server_profile.SPServer(data=data)


class ScanEnd(plg.Worker):
    """Complement the info after the scan is finished.
    """

    name = "scanend"
    descr = "conclude the scan"
    prio = 1000

    def run(self):
        """The entry point for the worker.
        """

        if self.client.server_issues:
            if not hasattr(self.server_profile, "server_malfunctions"):
                self.server_profile.server_malfunctions = []

            for malfunction in self.client.server_issues:
                self.server_profile.server_malfunctions.append(
                    server_profile.SPServerMalfunction(malfunction=malfunction)
                )

        scan_info = self.server_profile.scan_info
        start_time = scan_info.start_timestamp
        stop_time = time.time()
        scan_info.stop_timestamp = stop_time
        scan_info.stop_date = datetime.datetime.fromtimestamp(int(stop_time))
        scan_info.run_time = float(f"{stop_time - start_time:.3f}")



# -*- coding: utf-8 -*-
"""Module scanning for compression support
"""
# import basic stuff

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanCompression(plg.Worker):
    name = "compression"
    descr = "scan for compression support"
    prio = 30

    def _compression(self, version):
        self.server_profile.allocate_features()
        features = self.server_profile.features
        if not hasattr(features, "compression"):
            features.compression = []

        values = self.server_profile.get_profile_values([version])
        self.client.init_profile(profile_values=values)
        comp_methods = tls.CompressionMethod.all()

        while comp_methods:
            self.client.profile.compression_methods = comp_methods
            server_hello = None
            with self.client.create_connection() as conn:
                conn.send(msg.ClientHello)
                server_hello = conn.wait(msg.ServerHello)

            if server_hello is None:
                break

            if server_hello.version is not version:
                break

            if server_hello.compression_method not in comp_methods:
                break

            comp_methods.remove(server_hello.compression_method)
            if server_hello.compression_method not in features.compression:
                features.compression.append(server_hello.compression_method)

    def run(self):
        for version in self.server_profile.get_versions(exclude=[tls.Version.SSL20]):
            self._compression(version)



# -*- coding: utf-8 -*-
"""Module defining constants for the TLS protocol
"""
# import basic stuff
import enum
from typing import List, TypeVar, Type, Optional, Union

# import own stuff

# import other stuff

_T = TypeVar("_T")


class ExtendedEnum(enum.Enum):
    """Class to extend the enum.Enum class by additional methods
    """

    @classmethod
    def val2enum(cls: Type[_T], value: int, alert_on_failure: bool = False) -> _T:
        """Class method to map a value to the corresponding enum.

        Args:
            value: The enum value which is used to map to an enum
            alert_on_failure: If set to True and the value is not a valid enum
                value, an :obj:`ServerMalfunction` exception will be raised.
                Defaults to False.

        Returns:
            The corresponding enum or None, if the mapping fails and
            `alert_on_failure` is set to False.

        Raises:
            ServerMalfunction: In case the given value is not a valid enum and
                `alert_on_failure` is True
        """

        # TODO: resolve type issue
        enum = cls._value2member_map_.get(value)  # type: ignore
        if (enum is None) and alert_on_failure:
            raise ServerMalfunction(ServerIssue.ILLEGAL_PARAMETER_VALUE)

        return enum

    @classmethod
    def str2enum(cls, name: str, alert_on_failure: bool = False) -> Optional[enum.Enum]:
        """Class method to map a string to the corresponding enum.

        Args:
            name (str): The name which must correspond to the name of the enum.
            alert_on_failure (bool, optional): If set to True and the name is
                not a valid enum value, an :obj:`ServerMalfunction` exception will be
                raised. Defaults to False.

        Returns:
            The corresponding enum or None, if the mapping fails and
            `alert_on_failure` is set to False.

        Raises:
            ServerMalfunction: In case the given name is not a valid enum and
                `alert_on_failure` is True
        """

        enum = cls._member_map_.get(name)  # type: ignore
        if (enum is None) and alert_on_failure:
            raise ValueError(f"Value {name} not defined for {cls}")

        return enum

    @classmethod
    def all(cls) -> List["ExtendedEnum"]:
        """Get all enum items

        Returns:
            list of all enum items defined for that enum
        """

        return list(cls.__members__.values())

    def __str__(self) -> str:
        """Use the name as a string representation

        Returns:
            str: the name of the enum item
        """

        return self.name


class ExtendedIntEnum(ExtendedEnum):
    """Class for comparable enums

    Note, that we us our own class, as we cannot overwrite the __str__ method of the
    enum.IntEnum call (__slots__!)
    """

    def __lt__(self, other):
        return self.value < getattr(other, "value", other)

    def __gt__(self, other):
        return self.value > getattr(other, "value", other)

    def __le__(self, other):
        return self.value <= getattr(other, "value", other)

    def __ge__(self, other):
        return self.value >= getattr(other, "value", other)


class Entity(ExtendedEnum):
    """Enum used to represent the entity of the TLS connection endpoint.
    """

    CLIENT = 0
    SERVER = 1


class Version(ExtendedIntEnum):
    """Enum for the TLS versions.

    The values defined correspond to the values used in PDUs.
    """

    SSL20 = 0x0200
    SSL30 = 0x0300
    TLS10 = 0x0301
    TLS11 = 0x0302
    TLS12 = 0x0303
    TLS13 = 0x0304

    @classmethod
    def tls_only(cls) -> List["Version"]:
        """Comfortable method to get all TLS versions 1.0 .. 1.3, excluding SSLv2/v3.
        """

        return [cls.TLS10, cls.TLS11, cls.TLS12, cls.TLS13]


class ContentType(ExtendedEnum):
    """Enum representing the TLS-Handshake protocols.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    CHANGE_CIPHER_SPEC = 20
    ALERT = 21
    HANDSHAKE = 22
    APPLICATION_DATA = 23
    HEARTBEAT = 24
    SSL2 = 256


class CompressionMethod(ExtendedEnum):
    """Enum representing the TLS compression methods.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    NULL = 0
    DEFLATE = 1


class Extension(ExtendedEnum):
    """Enum representing the TLS extensions.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    SERVER_NAME = 0
    MAX_FRAGMENT_LENGTH = 1
    CLIENT_CERTIFICATE_URL = 2
    TRUSTED_CA_KEYS = 3
    TRUNCATED_HMAC = 4
    STATUS_REQUEST = 5
    USER_MAPPING = 6
    CLIENT_AUTHZ = 7
    SERVER_AUTHZ = 8
    CERT_TYPE = 9
    SUPPORTED_GROUPS = 10
    EC_POINT_FORMATS = 11
    SRP = 12
    SIGNATURE_ALGORITHMS = 13
    USE_SRTP = 14
    HEARTBEAT = 15
    APPLICATION_LAYER_PROTOCOL_NEGOTIATION = 16
    STATUS_REQUEST_V2 = 17
    SIGNED_CERTIFICATE_TIMESTAMP = 18
    CLIENT_CERTIFICATE_TYPE = 19
    SERVER_CERTIFICATE_TYPE = 20
    PADDING = 21
    ENCRYPT_THEN_MAC = 22
    EXTENDED_MASTER_SECRET = 23
    TOKEN_BINDING = 24
    CACHED_INFO = 25
    TLS_LTS = 26
    COMPRESS_CERTIFICATE = 27
    RECORD_SIZE_LIMIT = 28
    PWD_PROTECT = 29
    PWD_CLEAR = 30
    PASSWORD_SALT = 31
    TICKET_PINNING = 32
    TLS_CERT_WITH_EXTERN_PSK = 33
    DELEGATED_CREDENTIALS = 34
    SESSION_TICKET = 35
    SUPPORTED_EKT_CIPHERS = 39
    PRE_SHARED_KEY = 41
    EARLY_DATA = 42
    SUPPORTED_VERSIONS = 43
    COOKIE = 44
    PSK_KEY_EXCHANGE_MODES = 45
    CERTIFICATE_AUTHORITIES = 47
    OID_FILTERS = 48
    POST_HANDSHAKE_AUTH = 49
    SIGNATURE_ALGORITHMS_CERT = 50
    KEY_SHARE = 51
    TRANSPARENCY_INFO = 52
    CONNECTION_ID = 53
    EXTERNAL_ID_HASH = 55
    EXTERNAL_SESSION_ID = 56
    RENEGOTIATION_INFO = 65281
    UNKNOW_EXTENSION = 0x10000


class CipherSuite(ExtendedEnum):
    """Enum representing the TLS cipher suites (excluding SSL20).

    The values defined correspond to the values used in PDUs as defined by IANA,
    excluding cipher suites registered for drafts.
    """

    TLS_NULL_WITH_NULL_NULL = 0x0000
    TLS_RSA_WITH_NULL_MD5 = 0x0001
    TLS_RSA_WITH_NULL_SHA = 0x0002
    TLS_RSA_EXPORT_WITH_RC4_40_MD5 = 0x0003
    TLS_RSA_WITH_RC4_128_MD5 = 0x0004
    TLS_RSA_WITH_RC4_128_SHA = 0x0005
    TLS_RSA_EXPORT_WITH_RC2_CBC_40_MD5 = 0x0006
    TLS_RSA_WITH_IDEA_CBC_SHA = 0x0007
    TLS_RSA_EXPORT_WITH_DES40_CBC_SHA = 0x0008
    TLS_RSA_WITH_DES_CBC_SHA = 0x0009
    TLS_RSA_WITH_3DES_EDE_CBC_SHA = 0x000A
    TLS_DH_DSS_EXPORT_WITH_DES40_CBC_SHA = 0x000B
    TLS_DH_DSS_WITH_DES_CBC_SHA = 0x000C
    TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA = 0x000D
    TLS_DH_RSA_EXPORT_WITH_DES40_CBC_SHA = 0x000E
    TLS_DH_RSA_WITH_DES_CBC_SHA = 0x000F
    TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA = 0x0010
    TLS_DHE_DSS_EXPORT_WITH_DES40_CBC_SHA = 0x0011
    TLS_DHE_DSS_WITH_DES_CBC_SHA = 0x0012
    TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA = 0x0013
    TLS_DHE_RSA_EXPORT_WITH_DES40_CBC_SHA = 0x0014
    TLS_DHE_RSA_WITH_DES_CBC_SHA = 0x0015
    TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA = 0x0016
    TLS_DH_ANON_EXPORT_WITH_RC4_40_MD5 = 0x0017
    TLS_DH_ANON_WITH_RC4_128_MD5 = 0x0018
    TLS_DH_ANON_EXPORT_WITH_DES40_CBC_SHA = 0x0019
    TLS_DH_ANON_WITH_DES_CBC_SHA = 0x001A
    TLS_DH_ANON_WITH_3DES_EDE_CBC_SHA = 0x001B
    TLS_KRB5_WITH_DES_CBC_SHA = 0x001E
    TLS_KRB5_WITH_3DES_EDE_CBC_SHA = 0x001F
    TLS_KRB5_WITH_RC4_128_SHA = 0x0020
    TLS_KRB5_WITH_IDEA_CBC_SHA = 0x0021
    TLS_KRB5_WITH_DES_CBC_MD5 = 0x0022
    TLS_KRB5_WITH_3DES_EDE_CBC_MD5 = 0x0023
    TLS_KRB5_WITH_RC4_128_MD5 = 0x0024
    TLS_KRB5_WITH_IDEA_CBC_MD5 = 0x0025
    TLS_KRB5_EXPORT_WITH_DES_CBC_40_SHA = 0x0026
    TLS_KRB5_EXPORT_WITH_RC2_CBC_40_SHA = 0x0027
    TLS_KRB5_EXPORT_WITH_RC4_40_SHA = 0x0028
    TLS_KRB5_EXPORT_WITH_DES_CBC_40_MD5 = 0x0029
    TLS_KRB5_EXPORT_WITH_RC2_CBC_40_MD5 = 0x002A
    TLS_KRB5_EXPORT_WITH_RC4_40_MD5 = 0x002B
    TLS_PSK_WITH_NULL_SHA = 0x002C
    TLS_DHE_PSK_WITH_NULL_SHA = 0x002D
    TLS_RSA_PSK_WITH_NULL_SHA = 0x002E
    TLS_RSA_WITH_AES_128_CBC_SHA = 0x002F
    TLS_DH_DSS_WITH_AES_128_CBC_SHA = 0x0030
    TLS_DH_RSA_WITH_AES_128_CBC_SHA = 0x0031
    TLS_DHE_DSS_WITH_AES_128_CBC_SHA = 0x0032
    TLS_DHE_RSA_WITH_AES_128_CBC_SHA = 0x0033
    TLS_DH_ANON_WITH_AES_128_CBC_SHA = 0x0034
    TLS_RSA_WITH_AES_256_CBC_SHA = 0x0035
    TLS_DH_DSS_WITH_AES_256_CBC_SHA = 0x0036
    TLS_DH_RSA_WITH_AES_256_CBC_SHA = 0x0037
    TLS_DHE_DSS_WITH_AES_256_CBC_SHA = 0x0038
    TLS_DHE_RSA_WITH_AES_256_CBC_SHA = 0x0039
    TLS_DH_ANON_WITH_AES_256_CBC_SHA = 0x003A
    TLS_RSA_WITH_NULL_SHA256 = 0x003B
    TLS_RSA_WITH_AES_128_CBC_SHA256 = 0x003C
    TLS_RSA_WITH_AES_256_CBC_SHA256 = 0x003D
    TLS_DH_DSS_WITH_AES_128_CBC_SHA256 = 0x003E
    TLS_DH_RSA_WITH_AES_128_CBC_SHA256 = 0x003F
    TLS_DHE_DSS_WITH_AES_128_CBC_SHA256 = 0x0040
    TLS_RSA_WITH_CAMELLIA_128_CBC_SHA = 0x0041
    TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA = 0x0042
    TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA = 0x0043
    TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA = 0x0044
    TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA = 0x0045
    TLS_DH_ANON_WITH_CAMELLIA_128_CBC_SHA = 0x0046
    TLS_DHE_RSA_WITH_AES_128_CBC_SHA256 = 0x0067
    TLS_DH_DSS_WITH_AES_256_CBC_SHA256 = 0x0068
    TLS_DH_RSA_WITH_AES_256_CBC_SHA256 = 0x0069
    TLS_DHE_DSS_WITH_AES_256_CBC_SHA256 = 0x006A
    TLS_DHE_RSA_WITH_AES_256_CBC_SHA256 = 0x006B
    TLS_DH_ANON_WITH_AES_128_CBC_SHA256 = 0x006C
    TLS_DH_ANON_WITH_AES_256_CBC_SHA256 = 0x006D
    TLS_RSA_WITH_CAMELLIA_256_CBC_SHA = 0x0084
    TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA = 0x0085
    TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA = 0x0086
    TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA = 0x0087
    TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA = 0x0088
    TLS_DH_ANON_WITH_CAMELLIA_256_CBC_SHA = 0x0089
    TLS_PSK_WITH_RC4_128_SHA = 0x008A
    TLS_PSK_WITH_3DES_EDE_CBC_SHA = 0x008B
    TLS_PSK_WITH_AES_128_CBC_SHA = 0x008C
    TLS_PSK_WITH_AES_256_CBC_SHA = 0x008D
    TLS_DHE_PSK_WITH_RC4_128_SHA = 0x008E
    TLS_DHE_PSK_WITH_3DES_EDE_CBC_SHA = 0x008F
    TLS_DHE_PSK_WITH_AES_128_CBC_SHA = 0x0090
    TLS_DHE_PSK_WITH_AES_256_CBC_SHA = 0x0091
    TLS_RSA_PSK_WITH_RC4_128_SHA = 0x0092
    TLS_RSA_PSK_WITH_3DES_EDE_CBC_SHA = 0x0093
    TLS_RSA_PSK_WITH_AES_128_CBC_SHA = 0x0094
    TLS_RSA_PSK_WITH_AES_256_CBC_SHA = 0x0095
    TLS_RSA_WITH_SEED_CBC_SHA = 0x0096
    TLS_DH_DSS_WITH_SEED_CBC_SHA = 0x0097
    TLS_DH_RSA_WITH_SEED_CBC_SHA = 0x0098
    TLS_DHE_DSS_WITH_SEED_CBC_SHA = 0x0099
    TLS_DHE_RSA_WITH_SEED_CBC_SHA = 0x009A
    TLS_DH_ANON_WITH_SEED_CBC_SHA = 0x009B
    TLS_RSA_WITH_AES_128_GCM_SHA256 = 0x009C
    TLS_RSA_WITH_AES_256_GCM_SHA384 = 0x009D
    TLS_DHE_RSA_WITH_AES_128_GCM_SHA256 = 0x009E
    TLS_DHE_RSA_WITH_AES_256_GCM_SHA384 = 0x009F
    TLS_DH_RSA_WITH_AES_128_GCM_SHA256 = 0x00A0
    TLS_DH_RSA_WITH_AES_256_GCM_SHA384 = 0x00A1
    TLS_DHE_DSS_WITH_AES_128_GCM_SHA256 = 0x00A2
    TLS_DHE_DSS_WITH_AES_256_GCM_SHA384 = 0x00A3
    TLS_DH_DSS_WITH_AES_128_GCM_SHA256 = 0x00A4
    TLS_DH_DSS_WITH_AES_256_GCM_SHA384 = 0x00A5
    TLS_DH_ANON_WITH_AES_128_GCM_SHA256 = 0x00A6
    TLS_DH_ANON_WITH_AES_256_GCM_SHA384 = 0x00A7
    TLS_PSK_WITH_AES_128_GCM_SHA256 = 0x00A8
    TLS_PSK_WITH_AES_256_GCM_SHA384 = 0x00A9
    TLS_DHE_PSK_WITH_AES_128_GCM_SHA256 = 0x00AA
    TLS_DHE_PSK_WITH_AES_256_GCM_SHA384 = 0x00AB
    TLS_RSA_PSK_WITH_AES_128_GCM_SHA256 = 0x00AC
    TLS_RSA_PSK_WITH_AES_256_GCM_SHA384 = 0x00AD
    TLS_PSK_WITH_AES_128_CBC_SHA256 = 0x00AE
    TLS_PSK_WITH_AES_256_CBC_SHA384 = 0x00AF
    TLS_PSK_WITH_NULL_SHA256 = 0x00B0
    TLS_PSK_WITH_NULL_SHA384 = 0x00B1
    TLS_DHE_PSK_WITH_AES_128_CBC_SHA256 = 0x00B2
    TLS_DHE_PSK_WITH_AES_256_CBC_SHA384 = 0x00B3
    TLS_DHE_PSK_WITH_NULL_SHA256 = 0x00B4
    TLS_DHE_PSK_WITH_NULL_SHA384 = 0x00B5
    TLS_RSA_PSK_WITH_AES_128_CBC_SHA256 = 0x00B6
    TLS_RSA_PSK_WITH_AES_256_CBC_SHA384 = 0x00B7
    TLS_RSA_PSK_WITH_NULL_SHA256 = 0x00B8
    TLS_RSA_PSK_WITH_NULL_SHA384 = 0x00B9
    TLS_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BA
    TLS_DH_DSS_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BB
    TLS_DH_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BC
    TLS_DHE_DSS_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BD
    TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BE
    TLS_DH_ANON_WITH_CAMELLIA_128_CBC_SHA256 = 0x00BF
    TLS_RSA_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C0
    TLS_DH_DSS_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C1
    TLS_DH_RSA_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C2
    TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C3
    TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C4
    TLS_DH_ANON_WITH_CAMELLIA_256_CBC_SHA256 = 0x00C5
    TLS_EMPTY_RENEGOTIATION_INFO_SCSV = 0x00FF
    TLS_AES_128_GCM_SHA256 = 0x1301
    TLS_AES_256_GCM_SHA384 = 0x1302
    TLS_CHACHA20_POLY1305_SHA256 = 0x1303
    TLS_AES_128_CCM_SHA256 = 0x1304
    TLS_AES_128_CCM_8_SHA256 = 0x1305
    TLS_FALLBACK_SCSV = 0x5600
    TLS_ECDH_ECDSA_WITH_NULL_SHA = 0xC001
    TLS_ECDH_ECDSA_WITH_RC4_128_SHA = 0xC002
    TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA = 0xC003
    TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA = 0xC004
    TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA = 0xC005
    TLS_ECDHE_ECDSA_WITH_NULL_SHA = 0xC006
    TLS_ECDHE_ECDSA_WITH_RC4_128_SHA = 0xC007
    TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA = 0xC008
    TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA = 0xC009
    TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA = 0xC00A
    TLS_ECDH_RSA_WITH_NULL_SHA = 0xC00B
    TLS_ECDH_RSA_WITH_RC4_128_SHA = 0xC00C
    TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA = 0xC00D
    TLS_ECDH_RSA_WITH_AES_128_CBC_SHA = 0xC00E
    TLS_ECDH_RSA_WITH_AES_256_CBC_SHA = 0xC00F
    TLS_ECDHE_RSA_WITH_NULL_SHA = 0xC010
    TLS_ECDHE_RSA_WITH_RC4_128_SHA = 0xC011
    TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA = 0xC012
    TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA = 0xC013
    TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA = 0xC014
    TLS_ECDH_ANON_WITH_NULL_SHA = 0xC015
    TLS_ECDH_ANON_WITH_RC4_128_SHA = 0xC016
    TLS_ECDH_ANON_WITH_3DES_EDE_CBC_SHA = 0xC017
    TLS_ECDH_ANON_WITH_AES_128_CBC_SHA = 0xC018
    TLS_ECDH_ANON_WITH_AES_256_CBC_SHA = 0xC019
    TLS_SRP_SHA_WITH_3DES_EDE_CBC_SHA = 0xC01A
    TLS_SRP_SHA_RSA_WITH_3DES_EDE_CBC_SHA = 0xC01B
    TLS_SRP_SHA_DSS_WITH_3DES_EDE_CBC_SHA = 0xC01C
    TLS_SRP_SHA_WITH_AES_128_CBC_SHA = 0xC01D
    TLS_SRP_SHA_RSA_WITH_AES_128_CBC_SHA = 0xC01E
    TLS_SRP_SHA_DSS_WITH_AES_128_CBC_SHA = 0xC01F
    TLS_SRP_SHA_WITH_AES_256_CBC_SHA = 0xC020
    TLS_SRP_SHA_RSA_WITH_AES_256_CBC_SHA = 0xC021
    TLS_SRP_SHA_DSS_WITH_AES_256_CBC_SHA = 0xC022
    TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256 = 0xC023
    TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA384 = 0xC024
    TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256 = 0xC025
    TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA384 = 0xC026
    TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256 = 0xC027
    TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384 = 0xC028
    TLS_ECDH_RSA_WITH_AES_128_CBC_SHA256 = 0xC029
    TLS_ECDH_RSA_WITH_AES_256_CBC_SHA384 = 0xC02A
    TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256 = 0xC02B
    TLS_ECDHE_ECDSA_WITH_AES_256_GCM_SHA384 = 0xC02C
    TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256 = 0xC02D
    TLS_ECDH_ECDSA_WITH_AES_256_GCM_SHA384 = 0xC02E
    TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256 = 0xC02F
    TLS_ECDHE_RSA_WITH_AES_256_GCM_SHA384 = 0xC030
    TLS_ECDH_RSA_WITH_AES_128_GCM_SHA256 = 0xC031
    TLS_ECDH_RSA_WITH_AES_256_GCM_SHA384 = 0xC032
    TLS_ECDHE_PSK_WITH_RC4_128_SHA = 0xC033
    TLS_ECDHE_PSK_WITH_3DES_EDE_CBC_SHA = 0xC034
    TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA = 0xC035
    TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA = 0xC036
    TLS_ECDHE_PSK_WITH_AES_128_CBC_SHA256 = 0xC037
    TLS_ECDHE_PSK_WITH_AES_256_CBC_SHA384 = 0xC038
    TLS_ECDHE_PSK_WITH_NULL_SHA = 0xC039
    TLS_ECDHE_PSK_WITH_NULL_SHA256 = 0xC03A
    TLS_ECDHE_PSK_WITH_NULL_SHA384 = 0xC03B
    TLS_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC03C
    TLS_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC03D
    TLS_DH_DSS_WITH_ARIA_128_CBC_SHA256 = 0xC03E
    TLS_DH_DSS_WITH_ARIA_256_CBC_SHA384 = 0xC03F
    TLS_DH_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC040
    TLS_DH_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC041
    TLS_DHE_DSS_WITH_ARIA_128_CBC_SHA256 = 0xC042
    TLS_DHE_DSS_WITH_ARIA_256_CBC_SHA384 = 0xC043
    TLS_DHE_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC044
    TLS_DHE_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC045
    TLS_DH_ANON_WITH_ARIA_128_CBC_SHA256 = 0xC046
    TLS_DH_ANON_WITH_ARIA_256_CBC_SHA384 = 0xC047
    TLS_ECDHE_ECDSA_WITH_ARIA_128_CBC_SHA256 = 0xC048
    TLS_ECDHE_ECDSA_WITH_ARIA_256_CBC_SHA384 = 0xC049
    TLS_ECDH_ECDSA_WITH_ARIA_128_CBC_SHA256 = 0xC04A
    TLS_ECDH_ECDSA_WITH_ARIA_256_CBC_SHA384 = 0xC04B
    TLS_ECDHE_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC04C
    TLS_ECDHE_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC04D
    TLS_ECDH_RSA_WITH_ARIA_128_CBC_SHA256 = 0xC04E
    TLS_ECDH_RSA_WITH_ARIA_256_CBC_SHA384 = 0xC04F
    TLS_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC050
    TLS_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC051
    TLS_DHE_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC052
    TLS_DHE_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC053
    TLS_DH_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC054
    TLS_DH_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC055
    TLS_DHE_DSS_WITH_ARIA_128_GCM_SHA256 = 0xC056
    TLS_DHE_DSS_WITH_ARIA_256_GCM_SHA384 = 0xC057
    TLS_DH_DSS_WITH_ARIA_128_GCM_SHA256 = 0xC058
    TLS_DH_DSS_WITH_ARIA_256_GCM_SHA384 = 0xC059
    TLS_DH_ANON_WITH_ARIA_128_GCM_SHA256 = 0xC05A
    TLS_DH_ANON_WITH_ARIA_256_GCM_SHA384 = 0xC05B
    TLS_ECDHE_ECDSA_WITH_ARIA_128_GCM_SHA256 = 0xC05C
    TLS_ECDHE_ECDSA_WITH_ARIA_256_GCM_SHA384 = 0xC05D
    TLS_ECDH_ECDSA_WITH_ARIA_128_GCM_SHA256 = 0xC05E
    TLS_ECDH_ECDSA_WITH_ARIA_256_GCM_SHA384 = 0xC05F
    TLS_ECDHE_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC060
    TLS_ECDHE_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC061
    TLS_ECDH_RSA_WITH_ARIA_128_GCM_SHA256 = 0xC062
    TLS_ECDH_RSA_WITH_ARIA_256_GCM_SHA384 = 0xC063
    TLS_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC064
    TLS_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC065
    TLS_DHE_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC066
    TLS_DHE_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC067
    TLS_RSA_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC068
    TLS_RSA_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC069
    TLS_PSK_WITH_ARIA_128_GCM_SHA256 = 0xC06A
    TLS_PSK_WITH_ARIA_256_GCM_SHA384 = 0xC06B
    TLS_DHE_PSK_WITH_ARIA_128_GCM_SHA256 = 0xC06C
    TLS_DHE_PSK_WITH_ARIA_256_GCM_SHA384 = 0xC06D
    TLS_RSA_PSK_WITH_ARIA_128_GCM_SHA256 = 0xC06E
    TLS_RSA_PSK_WITH_ARIA_256_GCM_SHA384 = 0xC06F
    TLS_ECDHE_PSK_WITH_ARIA_128_CBC_SHA256 = 0xC070
    TLS_ECDHE_PSK_WITH_ARIA_256_CBC_SHA384 = 0xC071
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC072
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC073
    TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC074
    TLS_ECDH_ECDSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC075
    TLS_ECDHE_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC076
    TLS_ECDHE_RSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC077
    TLS_ECDH_RSA_WITH_CAMELLIA_128_CBC_SHA256 = 0xC078
    TLS_ECDH_RSA_WITH_CAMELLIA_256_CBC_SHA384 = 0xC079
    TLS_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC07A
    TLS_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC07B
    TLS_DHE_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC07C
    TLS_DHE_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC07D
    TLS_DH_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC07E
    TLS_DH_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC07F
    TLS_DHE_DSS_WITH_CAMELLIA_128_GCM_SHA256 = 0xC080
    TLS_DHE_DSS_WITH_CAMELLIA_256_GCM_SHA384 = 0xC081
    TLS_DH_DSS_WITH_CAMELLIA_128_GCM_SHA256 = 0xC082
    TLS_DH_DSS_WITH_CAMELLIA_256_GCM_SHA384 = 0xC083
    TLS_DH_ANON_WITH_CAMELLIA_128_GCM_SHA256 = 0xC084
    TLS_DH_ANON_WITH_CAMELLIA_256_GCM_SHA384 = 0xC085
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC086
    TLS_ECDHE_ECDSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC087
    TLS_ECDH_ECDSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC088
    TLS_ECDH_ECDSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC089
    TLS_ECDHE_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC08A
    TLS_ECDHE_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC08B
    TLS_ECDH_RSA_WITH_CAMELLIA_128_GCM_SHA256 = 0xC08C
    TLS_ECDH_RSA_WITH_CAMELLIA_256_GCM_SHA384 = 0xC08D
    TLS_PSK_WITH_CAMELLIA_128_GCM_SHA256 = 0xC08E
    TLS_PSK_WITH_CAMELLIA_256_GCM_SHA384 = 0xC08F
    TLS_DHE_PSK_WITH_CAMELLIA_128_GCM_SHA256 = 0xC090
    TLS_DHE_PSK_WITH_CAMELLIA_256_GCM_SHA384 = 0xC091
    TLS_RSA_PSK_WITH_CAMELLIA_128_GCM_SHA256 = 0xC092
    TLS_RSA_PSK_WITH_CAMELLIA_256_GCM_SHA384 = 0xC093
    TLS_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC094
    TLS_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC095
    TLS_DHE_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC096
    TLS_DHE_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC097
    TLS_RSA_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC098
    TLS_RSA_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC099
    TLS_ECDHE_PSK_WITH_CAMELLIA_128_CBC_SHA256 = 0xC09A
    TLS_ECDHE_PSK_WITH_CAMELLIA_256_CBC_SHA384 = 0xC09B
    TLS_RSA_WITH_AES_128_CCM = 0xC09C
    TLS_RSA_WITH_AES_256_CCM = 0xC09D
    TLS_DHE_RSA_WITH_AES_128_CCM = 0xC09E
    TLS_DHE_RSA_WITH_AES_256_CCM = 0xC09F
    TLS_RSA_WITH_AES_128_CCM_8 = 0xC0A0
    TLS_RSA_WITH_AES_256_CCM_8 = 0xC0A1
    TLS_DHE_RSA_WITH_AES_128_CCM_8 = 0xC0A2
    TLS_DHE_RSA_WITH_AES_256_CCM_8 = 0xC0A3
    TLS_PSK_WITH_AES_128_CCM = 0xC0A4
    TLS_PSK_WITH_AES_256_CCM = 0xC0A5
    TLS_DHE_PSK_WITH_AES_128_CCM = 0xC0A6
    TLS_DHE_PSK_WITH_AES_256_CCM = 0xC0A7
    TLS_PSK_WITH_AES_128_CCM_8 = 0xC0A8
    TLS_PSK_WITH_AES_256_CCM_8 = 0xC0A9
    TLS_PSK_DHE_WITH_AES_128_CCM_8 = 0xC0AA
    TLS_PSK_DHE_WITH_AES_256_CCM_8 = 0xC0AB
    TLS_ECDHE_ECDSA_WITH_AES_128_CCM = 0xC0AC
    TLS_ECDHE_ECDSA_WITH_AES_256_CCM = 0xC0AD
    TLS_ECDHE_ECDSA_WITH_AES_128_CCM_8 = 0xC0AE
    TLS_ECDHE_ECDSA_WITH_AES_256_CCM_8 = 0xC0AF
    TLS_ECCPWD_WITH_AES_128_GCM_SHA256 = 0xC0B0
    TLS_ECCPWD_WITH_AES_256_GCM_SHA384 = 0xC0B1
    TLS_ECCPWD_WITH_AES_128_CCM_SHA256 = 0xC0B2
    TLS_ECCPWD_WITH_AES_256_CCM_SHA384 = 0xC0B3
    TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256 = 0xCCA8
    TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256 = 0xCCA9
    TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAA
    TLS_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAB
    TLS_ECDHE_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAC
    TLS_DHE_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAD
    TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256 = 0xCCAE
    TLS_ECDHE_PSK_WITH_AES_128_GCM_SHA256 = 0xD001
    TLS_ECDHE_PSK_WITH_AES_256_GCM_SHA384 = 0xD002
    TLS_ECDHE_PSK_WITH_AES_128_CCM_8_SHA256 = 0xD003
    TLS_ECDHE_PSK_WITH_AES_128_CCM_SHA256 = 0xD005


class AlertDescription(ExtendedEnum):
    """Enum representing the alert descriptions.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    CLOSE_NOTIFY = 0
    UNEXPECTED_MESSAGE = 10
    BAD_RECORD_MAC = 20
    DECRYPTION_FAILED = 21
    RECORD_OVERFLOW = 22
    DECOMPRESSION_FAILED = 30
    HANDSHAKE_FAILURE = 40
    NO_CERTIFICATE = 41
    BAD_CERTIFICATE = 42
    UNSUPPORTED_CERTIFICATE = 43
    CERTIFICATE_REVOKED = 44
    CERTIFICATE_EXPIRED = 45
    CERTIFICATE_UNKNOWN = 46
    ILLEGAL_PARAMETER = 47
    UNKNOWN_CA = 48
    ACCESS_DENIED = 49
    DECODE_ERROR = 50
    DECRYPT_ERROR = 51
    EXPORT_RESTRICTION = 60
    PROTOCOL_VERSION = 70
    INSUFFICIENT_SECURITY = 71
    INTERNAL_ERROR = 80
    INAPPROPRIATE_FALLBACK = 86
    USER_CANCELED = 90
    NO_RENEGOTIATION = 100
    MISSING_EXTENSION = 109
    UNSUPPORTED_EXTENSION = 110
    CERTIFICATE_UNOBTAINABLE = 111
    UNRECOGNIZED_NAME = 112
    BAD_CERTIFICATE_STATUS_RESPONSE = 113
    BAD_CERTIFICATE_HASH_VALUE = 114
    UNKNOWN_PSK_IDENTITY = 115
    CERTIFICATE_REQUIRED = 116
    NO_APPLICATION_PROTOCOL = 120


class AlertLevel(ExtendedEnum):
    """Enum representing the alert level.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    WARNING = 1
    FATAL = 2


class HandshakeType(ExtendedEnum):
    """Enum representing the handshake message types.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    HELLO_REQUEST = 0
    CLIENT_HELLO = 1
    SERVER_HELLO = 2
    NEW_SESSION_TICKET = 4
    END_OF_EARLY_DATA = 5
    ENCRYPTED_EXTENSIONS = 8
    CERTIFICATE = 11
    SERVER_KEY_EXCHANGE = 12
    CERTIFICATE_REQUEST = 13
    SERVER_HELLO_DONE = 14
    CERTIFICATE_VERIFY = 15
    CLIENT_KEY_EXCHANGE = 16
    FINISHED = 20
    CERTIFICATE_STATUS = 22
    KEY_UPDATE = 24
    COMPRESSED_CERTIFICATE = 25
    EKT_KEY = 26
    MESSAGE_HASH = 254
    # HelloRetryRequest is identical to a ServerHello, we assign a value > 256
    # to it for a clear separation
    HELLO_RETRY_REQUEST = 258


class CCSType(ExtendedEnum):
    """Enum representing the message types for the Change Cipher Spec protocol.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    CHANGE_CIPHER_SPEC = 1


class HeartbeatType(ExtendedEnum):
    """Enum representing the message types for the Heartbeat protocol.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    HEARTBEAT_REQUEST = 1
    HEARTBEAT_RESPONSE = 2


class SupportedGroups(ExtendedEnum):
    """Enum representing the supported groups.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    SECT163K1 = 1
    SECT163R1 = 2
    SECT163R2 = 3
    SECT193R1 = 4
    SECT193R2 = 5
    SECT233K1 = 6
    SECT233R1 = 7
    SECT239K1 = 8
    SECT283K1 = 9
    SECT283R1 = 10
    SECT409K1 = 11
    SECT409R1 = 12
    SECT571K1 = 13
    SECT571R1 = 14
    SECP160K1 = 15
    SECP160R1 = 16
    SECP160R2 = 17
    SECP192K1 = 18
    SECP192R1 = 19
    SECP224K1 = 20
    SECP224R1 = 21
    SECP256K1 = 22
    SECP256R1 = 23
    SECP384R1 = 24
    SECP521R1 = 25
    BRAINPOOLP256R1 = 26
    BRAINPOOLP384R1 = 27
    BRAINPOOLP512R1 = 28
    X25519 = 29
    X448 = 30
    BRAINPOOLP256R1TLS13 = 31
    BRAINPOOLP384R1TLS13 = 32
    BRAINPOOLP512R1TLS13 = 33
    GC256A = 34
    GC256B = 35
    GC256C = 36
    GC256D = 37
    GC512A = 38
    GC512B = 39
    GC512C = 40
    CURVESM2 = 41
    FFDHE2048 = 256
    FFDHE3072 = 257
    FFDHE4096 = 258
    FFDHE6144 = 259
    FFDHE8192 = 260
    ARBITRARY_EXPLICIT_PRIME_CURVES = 65281
    ARBITRARY_EXPLICIT_CHAR2_CURVES = 65282

    @classmethod
    def all_tls13(cls) -> List["SupportedGroups"]:
        """Get all supported groups defined for TLS1.3 (RFC8446, 4.2.7)

        Returns:
            list (:obj:`SupportedGroups`): list of TLS1.3 supported groups
        """

        return [
            cls.SECP256R1,
            cls.SECP384R1,
            cls.SECP521R1,
            cls.X25519,
            cls.X448,
            cls.FFDHE2048,
            cls.FFDHE3072,
            cls.FFDHE4096,
            cls.FFDHE6144,
            cls.FFDHE8192,
        ]


class SignatureAlgorithm(ExtendedEnum):
    """Enum representing the signature algorithms.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    ANONYMOUS = 0
    RSA = 1
    DSA = 2
    ECDSA = 3
    ED25519 = 7
    ED448 = 8


class HashPrimitive(ExtendedEnum):
    """Enum representing the hash primitives.

    The values defined correspond to the values used in PDUs as defined by IANA,
    e.g. for the signature algorithms. But we use this enum generally, i.e., as well
    for the hash primitive given in the cipher suite.
    """

    NULL = 0
    MD5 = 1
    SHA1 = 2
    SHA224 = 3
    SHA256 = 4
    SHA384 = 5
    SHA512 = 6
    INTRINSIC = 8


class SignatureScheme(ExtendedEnum):
    """Enum representing the signature schemes.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    RSA_PKCS1_SHA1 = 0x0201
    ECDSA_SHA1 = 0x0203
    RSA_PKCS1_SHA256 = 0x0401
    ECDSA_SECP256R1_SHA256 = 0x0403
    RSA_PKCS1_SHA384 = 0x0501
    ECDSA_SECP384R1_SHA384 = 0x0503
    RSA_PKCS1_SHA512 = 0x0601
    ECDSA_SECP521R1_SHA512 = 0x0603
    ECCSI_SHA256 = 0x0704
    RSA_PSS_RSAE_SHA256 = 0x0804
    RSA_PSS_RSAE_SHA384 = 0x0805
    RSA_PSS_RSAE_SHA512 = 0x0806
    ED25519 = 0x0807
    ED448 = 0x0808
    RSA_PSS_PSS_SHA256 = 0x0809
    RSA_PSS_PSS_SHA384 = 0x080A
    RSA_PSS_PSS_SHA512 = 0x080B
    ECDSA_BRAINPOOLP256R1TLS13_SHA256 = 0x081A
    ECDSA_BRAINPOOLP384R1TLS13_SHA384 = 0x081B
    ECDSA_BRAINPOOLP512R1TLS13_SHA512 = 0x081C

    # ***************************************
    # legacy signature schemes needed as well
    # ***************************************
    RSA_PKCS1_MD5 = 0x0101
    RSA_PKCS1_SHA224 = 0x0301
    DSA_MD5 = 0x0102
    DSA_SHA1 = 0x0202
    DSA_SHA224 = 0x0302
    DSA_SHA256 = 0x0402
    DSA_SHA384 = 0x0502
    DSA_SHA512 = 0x0602
    ECDSA_SECP224R1_SHA224 = 0x0303


class EcPointFormat(ExtendedEnum):
    """Enum representing the Elliptic Curve Point Formats.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    UNCOMPRESSED = 0
    ANSIX962_COMPRESSED_PRIME = 1
    ANSIX962_COMPRESSED_CHAR2 = 2


class EcCurveType(ExtendedEnum):
    """Enum representing the curve type as received in the ServerKeyExchange message.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    EXPLICIT_PRIME = 1
    EXPLICIT_CHAR2 = 2
    NAMED_CURVE = 3


class KeyExchangeAlgorithm(ExtendedEnum):
    """Enum representing the key exchange mechanisms.
    """

    DHE_DSS = enum.auto()
    DHE_DSS_EXPORT = enum.auto()
    DHE_PSK = enum.auto()
    DHE_RSA = enum.auto()
    DHE_RSA_EXPORT = enum.auto()
    DH_ANON = enum.auto()
    DH_ANON_EXPORT = enum.auto()
    DH_DSS = enum.auto()
    DH_DSS_EXPORT = enum.auto()
    DH_RSA = enum.auto()
    DH_RSA_EXPORT = enum.auto()
    ECCPWD = enum.auto()
    ECDHE_ECDSA = enum.auto()
    ECDHE_PSK = enum.auto()
    ECDHE_RSA = enum.auto()
    ECDH_ANON = enum.auto()
    ECDH_ECDSA = enum.auto()
    ECDH_RSA = enum.auto()
    KRB5 = enum.auto()
    KRB5_EXPORT = enum.auto()
    NULL = enum.auto()
    PSK = enum.auto()
    PSK_DHE = enum.auto()
    RSA = enum.auto()
    RSA_EXPORT = enum.auto()
    RSA_PSK = enum.auto()
    SRP_SHA = enum.auto()
    SRP_SHA_DSS = enum.auto()
    SRP_SHA_RSA = enum.auto()
    TLS13_KEY_SHARE = enum.auto()


class KeyExchangeType(ExtendedEnum):
    """Enum representing the key exchange types.
    """

    NONE = enum.auto()
    RSA = enum.auto()
    DH = enum.auto()
    ECDH = enum.auto()


class KeyAuthentication(ExtendedEnum):
    """Enum representing the key authentication method.
    """

    NONE = enum.auto()
    RSA = enum.auto()
    DSS = enum.auto()
    ECDSA = enum.auto()


class CipherPrimitive(ExtendedEnum):
    """Enum representing the cipher primitive.
    """

    NULL = enum.auto()
    AES = enum.auto()
    CAMELLIA = enum.auto()
    IDEA = enum.auto()
    ARIA = enum.auto()
    RC4 = enum.auto()
    RC2 = enum.auto()
    SEED = enum.auto()
    DES = enum.auto()
    TRIPPLE_DES = enum.auto()
    CHACHA = enum.auto()


class SymmetricCipher(ExtendedEnum):
    """Enum representing the ciphers supported by tlsmate.

    "Supported" means, a handshake can be completed successfully, and a secured
    channel is established between both peers.
    """

    TRIPPLE_DES_EDE_CBC = enum.auto()
    AES_128 = enum.auto()
    AES_128_CBC = enum.auto()
    AES_128_CCM = enum.auto()
    AES_128_CCM_8 = enum.auto()
    AES_128_GCM = enum.auto()
    AES_256 = enum.auto()
    AES_256_CBC = enum.auto()
    AES_256_CCM = enum.auto()
    AES_256_CCM_8 = enum.auto()
    AES_256_GCM = enum.auto()
    ARIA_128_CBC = enum.auto()
    ARIA_128_GCM = enum.auto()
    ARIA_256_CBC = enum.auto()
    ARIA_256_GCM = enum.auto()
    CAMELLIA_128_CBC = enum.auto()
    CAMELLIA_128_GCM = enum.auto()
    CAMELLIA_256_CBC = enum.auto()
    CAMELLIA_256_GCM = enum.auto()
    CHACHA20_POLY1305 = enum.auto()
    DES40_CBC = enum.auto()
    DES_CBC = enum.auto()
    DES_CBC_40 = enum.auto()
    IDEA_CBC = enum.auto()
    NULL = enum.auto()
    RC2_CBC_40 = enum.auto()
    RC4_128 = enum.auto()
    RC4_40 = enum.auto()
    SEED_CBC = enum.auto()
    TLS13_AES_128_GCM = enum.auto()
    TLS13_AES_256_GCM = enum.auto()
    TLS13_AES_128_CCM = enum.auto()
    TLS13_AES_128_CCM_8 = enum.auto()


class CipherType(ExtendedEnum):
    """Enum representing the type of the cipher.
    """

    NULL = enum.auto()
    BLOCK = enum.auto()
    STREAM = enum.auto()
    AEAD = enum.auto()


class PskKeyExchangeMode(ExtendedEnum):
    """Values used in the extension psk_key_exchange_mode.
    """

    PSK_KE = 0
    PSK_DHE_KE = 1


class HeartbeatMode(ExtendedEnum):
    """Values for the heartbeat mode (extension)
    """

    PEER_ALLOWED_TO_SEND = 1
    PEER_NOT_ALLOWED_TO_SEND = 2


class ScanState(ExtendedEnum):
    """Enum representing a pseudo-boolean value in the server profile.

    In addition to True and False the two values are defined:

    * UNDETERMINED - used when tlsmate has not even tried to determine the
      value or if the value could not determined at all (for whatever reason).

    * NA - Used as an indication this the boolean value is not applicable, e.g.
      when the server does not support any CBC-cipher suite, support for the
      extension ENCRYPT_THEN_MAC is not applicable.
    """

    FALSE = 0
    TRUE = 1
    NA = 2
    UNDETERMINED = 3


class SSLMessagType(ExtendedEnum):
    """Enum representing the message types for SSL2.

    The values defined correspond to the values used in PDUs as defined by IANA.
    """

    SSL2_ERROR = 0
    SSL2_CLIENT_HELLO = 1
    SSL2_CLIENT_MASTER_KEY = 2
    SSL2_CLIENT_FINISHED = 3
    SSL2_SERVER_HELLO = 4
    SSL2_SERVER_VERIFY = 5
    SSL2_SERVER_FINISHED = 6
    SSL2_REQUEST_CERTIFICATE = 7
    SSL2_CLIENT_CERTIFICATE = 8


class SSLCipherKind(ExtendedEnum):
    """Enum representing cipher suite ("cipher kind" in SSL2 jargon).

    The values defined correspond to the values used in PDUs.

    Note:
        The cipher kind has a length of 3 bytes and is completely disjunct from the
        TLS cipher suites.
    """

    SSL_CK_RC4_128_WITH_MD5 = 0x010080
    SSL_CK_RC4_128_EXPORT40_WITH_MD5 = 0x020080
    SSL_CK_RC2_128_CBC_WITH_MD5 = 0x030080
    SSL_CK_RC2_128_CBC_EXPORT40_WITH_MD5 = 0x040080
    SSL_CK_IDEA_128_CBC_WITH_MD5 = 0x050080
    SSL_CK_DES_64_CBC_WITH_MD5 = 0x060040
    SSL_CK_DES_192_EDE3_CBC_WITH_MD5 = 0x0700C0


class SSLVersion(ExtendedEnum):
    """Enum representing the SSL2 version.

    The values defined correspond to the values used in PDUs.
    """

    SSL2 = 0x0002


class SSLError(ExtendedEnum):
    """Enum representing SSLv2 errors.
    """

    NO_CIPHER = 0x0001
    NO_CERTIFICATE = 0x0002
    BAD_CERTIFICATE = 0x0004
    UNSUPPORTED_CERTIFICATE_TYPE = 0x0006


class CertType(ExtendedEnum):
    """Enum representing certificate types.
    """

    RSA_SIGN = 1
    DSS_SIGN = 2
    RSA_FIXED_DH = 3
    DSS_FIXED_DH = 4
    RSA_EPHEMERAL_DH = 5
    DSS_EPHEMERAL_DH = 6
    FORTEZZA_DMS = 20
    ECDSA_SIGN = 64
    RSA_FIXED_ECDH = 65
    ECDSA_FIXED_ECDH = 66


class CertKeyUsage(ExtendedEnum):
    """Representing the various flags of the KeyUsage extension for certificates
    """

    DIGITAL_SIGNATURE = enum.auto()
    CONTENT_COMMITMENT = enum.auto()
    KEY_ENCIPHERMENT = enum.auto()
    DATA_ENCIPHERMENT = enum.auto()
    KEY_AGREEMENT = enum.auto()
    KEY_CERT_SIGN = enum.auto()
    CRL_SIGN = enum.auto()
    ENCIPHER_ONLY = enum.auto()
    DECIPHER_ONLY = enum.auto()


class CertCrlStatus(ExtendedEnum):
    """CRL Revocation Status
    """

    UNDETERMINED = enum.auto()
    NOT_REVOKED = enum.auto()
    REVOKED = enum.auto()
    CRL_DOWNLOAD_FAILED = enum.auto()
    WRONG_CRL_ISSUER = enum.auto()
    INVALID_TIMESTAMP = enum.auto()
    CRL_SIGNATURE_INVALID = enum.auto()


class OcspStatus(ExtendedEnum):
    """OCSP revocation status
    """

    NOT_APPLICABLE = enum.auto()
    NOT_SUPPORTED = enum.auto()
    UNDETERMINED = enum.auto()
    NOT_REVOKED = enum.auto()
    REVOKED = enum.auto()
    UNKNOWN = enum.auto()
    TIMEOUT = enum.auto()
    INVALID_RESPONSE = enum.auto()
    SIGNATURE_INVALID = enum.auto()
    INVALID_TIMESTAMP = enum.auto()
    NO_ISSUER = enum.auto()
    INVALID_ISSUER_CERT = enum.auto()


class Profile(ExtendedEnum):
    """Different types for client profile
    """

    LEGACY = enum.auto()
    INTEROPERABILITY = enum.auto()
    MODERN = enum.auto()
    TLS13 = enum.auto()


class HostType(ExtendedEnum):
    """Type of an IP-endpoint
    """

    HOST = enum.auto()
    IPV4 = enum.auto()
    IPV6 = enum.auto()


class RobotVulnerability(ExtendedEnum):
    """Status for the ROBOT vulnerability
    """

    NOT_APPLICABLE = enum.auto()
    UNDETERMINED = enum.auto()
    INCONSITENT_RESULTS = enum.auto()
    WEAK_ORACLE = enum.auto()
    STRONG_ORACLE = enum.auto()
    NOT_VULNERABLE = enum.auto()


class HeartbleedStatus(ExtendedEnum):
    """Status for heartbleed vulnerability
    """

    NOT_APPLICABLE = enum.auto()
    UNDETERMINED = enum.auto()
    VULNERABLE = enum.auto()
    NOT_VULNERABLE = enum.auto()
    TIMEOUT = enum.auto()
    CONNECTION_CLOSED = enum.auto()


class HeartbeatState(ExtendedEnum):
    """Status for heartbeart support
    """

    FALSE = 0
    TRUE = 1
    NA = 2
    UNDETERMINED = 3
    NOT_REPONDING = 4
    WRONG_RESPONSE = 5
    UNEXPECTED_MESSAGE = 6


class StatusType(ExtendedEnum):
    """Status type for TLS extension status_request
    """

    OCSP = 1
    OCSP_MULTI = 2
    NONE = 256


class SPCbcPaddingOracle(ExtendedEnum):
    """Different types of CBC padding oracles
    """

    LUCKY_MINUS_20 = enum.auto()
    PADDING_FILLS_RECORD = enum.auto()
    PADDING_EXCEEDS_RECORD = enum.auto()
    INVALID_PADDING = enum.auto()
    INVALID_MAC = enum.auto()


class OracleScanAccuracy(ExtendedEnum):
    """How accurate the scan for CBC padding oracles shall be
    """

    LOW = enum.auto()
    MEDIUM = enum.auto()
    HIGH = enum.auto()


class ServerIssue(ExtendedEnum):
    """Indication of a severe server violation
    """

    PSK_OUT_OF_RANGE = "selected PSK out of range (TLS1.3)"
    KEY_SHARE_NOT_PRESENT = "ServerHello, TLS13: extension KEY_SHARE not present"
    SECURE_RENEG_FAILED = "secure renegotiation check failed"
    VERIFY_DATA_INVALID = "received Finished: verify data does not match"
    CERT_REQ_NO_SIG_ALGO = (
        "certificate request without extension SignatureAlgorithms received"
    )
    EXTENTION_LENGHT_ERROR = "extension length incorrect"
    SNI_NO_HOSTNAME = "host_name not present"
    FFDH_GROUP_UNKNOWN = "FF-DH group unknown"
    MESSAGE_LENGTH_ERROR = "message length incorrect"
    INCOMPATIBLE_KEY_EXCHANGE = (
        "key exchange algorithm in ServerKeyExchange message incompatible with "
        "offered cipher suite"
    )
    PARAMETER_LENGTH_ERROR = "message length error when unpacking parameter"
    RECORD_TOO_SHORT = "decoded record shorter than MAC length"
    RECORD_MAC_INVALID = "MAC verification failed"
    RECORD_WRONG_PADDING_LENGTH = "wrong padding length"
    RECORD_WRONG_PADDING_BYTES = "wrong padding byte contents"
    ILLEGAL_PARAMETER_VALUE = "received parameter value is illegal"
    KEX_INVALID_SIGNATURE = "signature of server's key exchange parameters invalid"


class Logjam(ExtendedEnum):
    """The type of logjam weakness
    """

    NA = enum.auto()
    OK = enum.auto()
    UNDETERMINED = enum.auto()
    PRIME512 = enum.auto()
    PRIME1024_COMMON = enum.auto()
    PRIME1024_CUSTOMIZED = enum.auto()


##############
# Exceptions #
##############


class TlsmateException(Exception):
    """A class all exception for tlsmate are based on.
    """


class ServerMalfunction(TlsmateException):
    """Exception raised in case the server response contains unrecoverable errors.

    This exception basically indicates a specification violation by the server.

    Attributes:
        issue: the reason for the exception
        message: the message, if applicable
        extension: the extension, if applicable
    """

    def __init__(
        self,
        issue: ServerIssue,
        message: Optional[Union[HandshakeType, CCSType]] = None,
        extension: Optional[Extension] = None,
    ) -> None:
        super().__init__(issue.value)
        self.issue = issue
        self.message = message
        self.extension = extension


class TlsConnectionClosedError(TlsmateException):
    """Exception raised when the TLS connection is closed unexpectedly.

    Attributes:
        exc: the original exception
    """

    def __init__(self, exc: Optional[Exception] = None) -> None:
        self.exc = exc


class TlsMsgTimeoutError(TlsmateException):
    """Exception raised when a message is not received within a given timeout.
    """

    pass


class CurveNotSupportedError(TlsmateException):
    """Exception raised when a curve is negotiated which is not supported.

    Attributes:
        message: A human readable string providing the cause
        curve: The curve has been offered by the client, and selected by the
            server, but it is not supported for a full key exchange.
    """

    def __init__(self, message: str, curve: SupportedGroups) -> None:
        self.message = message
        self.curve = curve


class ScanError(TlsmateException):
    """Exception which might occur during a scan.

    The exception will be raised if an abnormal condition during a scan is
    detected.

    Attributes:
        message: A human readable string describing the cause.
    """

    def __init__(self, message: str) -> None:
        self.message = message


class OcspError(TlsmateException):
    """Exception for OCSP errors

    Attributes:
        issue: A human readable string describing the cause.
    """

    def __init__(self, issue: str) -> None:
        self.issue = issue


class UntrustedCertificate(TlsmateException):
    """Exception for unsuccessful certificate (chain) validation.

    Attributes:
        issue: A human readable string describing the cause.
    """

    def __init__(self, issue: str) -> None:
        self.issue = issue


class ServerParmsSignatureInvalid(TlsmateException):
    """More user friendly exception than cryptography.exception.InvalidSignature
    """



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.sig_algo import ScanSigAlgs
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary

sig_algs = [
    "DSA_SHA256",
    "DSA_SHA1",
]


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_supported_groups_openssl1_0_2_DSA"
    sp_out_yaml = "profile_sig_algos_openssl1_0_2_DSA"
    recorder_yaml = "recorder_sig_algos_openssl1_0_2_DSA"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-dsa "
        "-- -www -cipher ALL"
    )
    library = TlsLibrary.openssl1_0_2

    server = "localhost"

    def check_profile(self, profile):
        assert "signature_algorithms" not in profile["versions"][1]  # SSL30
        assert "signature_algorithms" not in profile["versions"][2]  # TLS10
        assert "signature_algorithms" not in profile["versions"][3]  # TLS11
        prof = profile["versions"][4]  # TLS12
        for a, b in zip(sig_algs, prof["signature_algorithms"]["algorithms"]):
            assert a == b["name"]

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        ScanSigAlgs(tlsmate).run()

        self.check_profile(server_profile.make_serializable())


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Implement unit tests for the module utils.
"""
import re
from tlsmate import utils
from tlsmate import tls

data = {
    "array": [1, 2, 3],
    "integer": 4,
    "object": {"a": 1, "b": 2},
    "string": "hello",
}

json = (
    "{\n"
    '    "array": [\n'
    "        1,\n"
    "        2,\n"
    "        3\n"
    "    ],\n"
    '    "integer": 4,\n'
    '    "object": {\n'
    '        "a": 1,\n'
    '        "b": 2\n'
    "    },\n"
    '    "string": "hello"\n'
    "}"
)

yaml = (
    "array:\n"
    "- 1\n"
    "- 2\n"
    "- 3\n"
    "integer: 4\n"
    "object:\n"
    "    a: 1\n"
    "    b: 2\n"
    "string: hello\n"
)


def test_get_random_value():
    val = utils.get_random_value()
    assert type(val) is bytearray
    assert len(val) == 32


def test_serialize_data_file(tmp_path):
    file_name = tmp_path / "output.txt"
    utils.serialize_data(data, file_name)
    content = file_name.read_text()
    assert content == yaml


def test_serialize_data_file_exists(tmp_path, capsys):
    file_name = tmp_path / "output.txt"
    file_name.touch()
    utils.serialize_data(data, file_name, replace=False)
    captured = capsys.readouterr()
    assert re.match(r"File .* existing\. .*-file not generated", captured.out)


def test_serialize_data_file_json(tmp_path):
    file_name = tmp_path / "output.txt"
    utils.serialize_data(data, file_name, use_json=True)
    content = file_name.read_text()
    assert content == json


def test_serialize_data_file_json_console(capsys):
    utils.serialize_data(data, use_json=True)
    captured = capsys.readouterr()
    assert captured.out == json + "\n"


def test_serialize_data_file_yaml_console(capsys):
    utils.serialize_data(data)
    captured = capsys.readouterr()
    assert captured.out == yaml + "\n"


def test_fold_string():
    string = "Hello world, this is a long string! Have fun!"
    assert utils.fold_string(string, 15) == [
        "Hello world, ",
        "this is a long ",
        "string! Have ",
        "fun!",
    ]


def test_filter_cipher_suites_key_exchange_supported():
    cipher_suites = [
        tls.CipherSuite.TLS_KRB5_WITH_DES_CBC_SHA,
        tls.CipherSuite.TLS_RSA_WITH_RC4_128_MD5,
    ]
    assert utils.filter_cipher_suites(cipher_suites, key_exchange_supported=True) == [
        tls.CipherSuite.TLS_RSA_WITH_RC4_128_MD5
    ]


def test_int_to_bytes():
    assert utils.int_to_bytes(0) == b"\0"
    assert utils.int_to_bytes(255) == b"\xff"
    assert utils.int_to_bytes(256) == b"\01\00"


def check_presence(cs_list, string_list):
    for cs in cs_list:
        match = False
        for string in string_list:
            if string in cs.name:
                match = True
                break
        assert match is True


def check_absence(cs_list, string):
    for cs in cs_list:
        assert string not in cs.name


def test_filter_cipher_suites():
    """Simple tests, only one filter condition at once
    """
    cs_all = tls.CipherSuite.all()
    cs_all.remove(tls.CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV)
    cs_all.remove(tls.CipherSuite.TLS_FALLBACK_SCSV)

    assert cs_all == utils.filter_cipher_suites(cs_all)

    tls13 = utils.filter_cipher_suites(cs_all, version=tls.Version.TLS13)
    assert set(tls13) == set(
        (
            tls.CipherSuite.TLS_AES_128_GCM_SHA256,
            tls.CipherSuite.TLS_AES_256_GCM_SHA384,
            tls.CipherSuite.TLS_CHACHA20_POLY1305_SHA256,
            tls.CipherSuite.TLS_AES_128_CCM_SHA256,
            tls.CipherSuite.TLS_AES_128_CCM_8_SHA256,
        )
    )

    tls12 = utils.filter_cipher_suites(cs_all, version=tls.Version.TLS12)
    assert set(tls12) & set(tls13) == set()
    assert set(tls12) | set(tls13) == set(cs_all)

    tls11 = utils.filter_cipher_suites(cs_all, version=tls.Version.TLS11)
    diff = set(tls12).difference(set(tls11))
    aead = utils.filter_cipher_suites(tls12, cipher_type=[tls.CipherType.AEAD])
    assert diff != set()
    assert set(aead) == set(diff)
    check_presence(aead, ["_GCM_", "_CCM", "_CHACHA20_POLY1305_"])
    check_absence(tls11, "_GCM_")
    check_absence(tls11, "_CCM")
    check_absence(tls11, "_CHACHA20_POLY1305_")

    ecdh_rsa = utils.filter_cipher_suites(
        cs_all, key_algo=[tls.KeyExchangeAlgorithm.ECDH_RSA], remove=True
    )
    check_presence(ecdh_rsa, ["_ECDH_RSA_"])
    check_absence(cs_all, "_ECDH_RSA_")

    cs_all = tls.CipherSuite.all()
    dh = utils.filter_cipher_suites(
        cs_all, key_exch=[tls.KeyExchangeType.DH], remove=True
    )
    check_presence(dh, ["_DH_", "_DHE_"])
    check_absence(cs_all, "_DH_")
    check_absence(cs_all, "_DHE_")

    cs_all = tls.CipherSuite.all()
    ecdsa = utils.filter_cipher_suites(
        cs_all, key_auth=[tls.KeyAuthentication.ECDSA], remove=True
    )
    check_presence(ecdsa, ["_ECDSA_"])
    check_absence(cs_all, "_ECDSA_")

    cs_all = tls.CipherSuite.all()
    ecdsa = utils.filter_cipher_suites(
        cs_all, cipher=[tls.SymmetricCipher.AES_128_CBC], remove=True
    )
    check_presence(ecdsa, ["_AES_128_CBC_"])
    check_absence(cs_all, "_AES_128_CBC_")

    cs_all = tls.CipherSuite.all()
    ecdsa = utils.filter_cipher_suites(
        cs_all, cipher_prim=[tls.CipherPrimitive.ARIA], remove=True
    )
    check_presence(ecdsa, ["_WITH_ARIA_"])
    check_absence(cs_all, "_WITH_ARIA_")

    cs_all = tls.CipherSuite.all()
    sha384 = utils.filter_cipher_suites(
        cs_all, mac=[tls.HashPrimitive.SHA384], remove=True
    )
    check_presence(sha384, ["_SHA384"])
    check_absence(cs_all, "_SHA384")

    cs_all = tls.CipherSuite.all()
    cs_all.remove(tls.CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV)
    cs_all.remove(tls.CipherSuite.TLS_FALLBACK_SCSV)
    cs_all2 = cs_all[:]
    full_hs = utils.filter_cipher_suites(cs_all, full_hs=True, remove=True)
    not_full_hs = utils.filter_cipher_suites(cs_all2, full_hs=False, remove=True)
    assert set(full_hs) != set()
    assert not_full_hs != set()
    assert cs_all == not_full_hs
    assert cs_all2 == full_hs



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tests.cipher_suite_tester import CipherSuiteTester
from tlsmate import tls
from tlsmate.tlssuite import TlsLibrary


class TestCase(CipherSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the CipherSuiteTester class.
    """

    path = pathlib.Path(__file__)

    cipher_suite = tls.CipherSuite.TLS_DHE_DSS_WITH_AES_128_CBC_SHA256
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-dsa "
        "-- -www -cipher ALL"
    )
    library = TlsLibrary.openssl1_1_1

    signature_algorithms = [
        tls.SignatureScheme.ECDSA_SECP256R1_SHA256,
        tls.SignatureScheme.ECDSA_SECP384R1_SHA384,
        tls.SignatureScheme.ECDSA_SECP521R1_SHA512,
        tls.SignatureScheme.RSA_PSS_RSAE_SHA256,
        tls.SignatureScheme.RSA_PSS_RSAE_SHA384,
        tls.SignatureScheme.RSA_PSS_RSAE_SHA512,
        tls.SignatureScheme.RSA_PKCS1_SHA256,
        tls.SignatureScheme.RSA_PKCS1_SHA384,
        tls.SignatureScheme.RSA_PKCS1_SHA512,
        tls.SignatureScheme.ECDSA_SHA1,
        tls.SignatureScheme.RSA_PKCS1_SHA1,
        tls.SignatureScheme.DSA_SHA1,
        tls.SignatureScheme.DSA_SHA256,
        tls.SignatureScheme.DSA_SHA384,
    ]
    # Uncomment the line below if you do not want to use the default version and
    # adapt it to your needs.
    # version = tls.Version.TLS12


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module containing classes for the key exchange
"""
# import basic stuff
import abc
import os
from typing import Optional, Any, Type, Union, NamedTuple

# import own stuff
import tlsmate.dh_numbers as dh_numbers
import tlsmate.pdu as pdu
import tlsmate.recorder as rec
import tlsmate.tls as tls

# import other stuff
from cryptography.hazmat.primitives.asymmetric import ec, x25519, x448, dh
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PublicFormat,
    PrivateFormat,
    NoEncryption,
    load_der_private_key,
)


class KeyExchange(metaclass=abc.ABCMeta):
    """The abstract class to derive different key exchange classes from
    """

    def __init__(self, recorder: rec.Recorder, **kwargs: Any) -> None:
        self._recorder = recorder

    @abc.abstractmethod
    def set_remote_key(self, rem_pub_key: bytes) -> None:
        """Store the remote public key
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_transferable_key(self):
        """Returns the pdu-format for whatever is transferred to the peer
        """

        raise NotImplementedError

    @abc.abstractmethod
    def get_shared_secret(self):
        """Do everything needed to provide the premaster secret
        """

        raise NotImplementedError

    def get_key_share(self) -> bytes:
        """To be implemented by classes which are applicable for TLS1.3
        """
        raise NotImplementedError

    def set_params(self, **kwargs: Any) -> None:
        """Sets parameter specific to the type of key exchange
        """

        pass


class RsaKeyExchange(KeyExchange):
    """Implement an RSA based key transport
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._pms: Optional[bytearray] = None
        self._version: Optional[tls.Version] = None
        self._rem_pub_key: Optional[rsa.RSAPublicKey] = None
        super().__init__(*args, **kwargs)

    def set_remote_key(self, rem_pub_key, **kwargs):
        """Not applicable for RSA based ley transport
        """

        raise NotImplementedError

    def _create_pms(self) -> None:
        """Build the premaster secret

        The client selects a random number, preceded by the TLS version as sent in
        the ClientHello.
        """

        assert self._version
        pms = bytearray()
        pms.extend(pdu.pack_uint16(self._version.value))
        random = self._recorder.inject(pms_rsa=os.urandom(46))
        pms.extend(random)
        self._pms = pms

    def get_transferable_key(self) -> bytes:
        """Returns an RSA-encrypted key

        Returns:
            the encrypted key
        """

        if self._pms is None:
            self._create_pms()
            assert self._pms

        assert self._rem_pub_key
        ciphered_key = self._rem_pub_key.encrypt(bytes(self._pms), padding.PKCS1v15())
        # injecting the encrypted key to the recorder is required, as the
        # padding scheme PKCS1v15 produces non-deterministic cipher text.
        return self._recorder.inject(rsa_enciphered=ciphered_key)

    def get_shared_secret(self) -> bytes:
        """Get the shared key, i.e., the premaster secret.

        Returns:
            the premaster secret
        """

        if self._pms is None:
            self._create_pms()
            assert self._pms

        return self._pms

    def set_params(
        self,
        version: Optional[tls.Version] = None,
        rem_public_key: Optional[rsa.RSAPublicKey] = None,
        **kwargs: Any,
    ) -> None:
        """Sets the versions and the remote public key
        """

        if version:
            self._version = version

        if rem_public_key:
            self._rem_pub_key = rem_public_key


class DhKeyExchange(KeyExchange):
    """Implement Diffie-Hellman key exchange
    """

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        self._pval: Optional[int] = None
        self._gval: Optional[int] = None
        self._rem_pub_key: Optional[int] = None
        self._priv_key = None
        self._pub_key = None
        self._dh_group = None

    def _create_key_pair(self):
        self._dh_group = dh.DHParameterNumbers(self._pval, self._gval)
        if self._recorder.is_injecting():
            x = self._recorder.inject(x_val=None)
            self._priv_key = load_der_private_key(x, None)
            self._pub_key = self._priv_key.public_key()

        else:
            self._priv_key = self._dh_group.parameters().generate_private_key()
            self._pub_key = self._priv_key.public_key()
            x = self._priv_key.private_bytes(
                Encoding.DER, PrivateFormat.PKCS8, NoEncryption()
            )
            self._recorder.trace(x_val=x)

    def get_shared_secret(self) -> bytes:
        """Get the shared key, i.e., the premaster secret.

        Returns:
            the premaster secret
        """

        if self._priv_key is None:
            self._create_key_pair()
            assert self._priv_key

        rem_pub_numbers = dh.DHPublicNumbers(self._rem_pub_key, self._dh_group)
        rem_pub_key = rem_pub_numbers.public_key()
        return self._priv_key.exchange(rem_pub_key).lstrip(b"\0")

    def set_remote_key(
        self,
        rem_pub_key: bytes,
        g_val: Optional[int] = None,
        p_val: Optional[bytes] = None,
        group: Optional[tls.SupportedGroups] = None,
    ) -> None:
        """Sets the remote public key

        Arguments:
            rem_pub_key: the raw remote public key
            g_val: the generator value
            p_val: the p-value as a byte string
            group: the supported group
        """

        if group is not None:
            dh_nbrs = dh_numbers.dh_numbers.get(group)
            if dh_nbrs is None:
                raise tls.ServerMalfunction(tls.ServerIssue.FFDH_GROUP_UNKNOWN)

            p_val = dh_nbrs.p_val
            g_val = dh_nbrs.g_val

        assert p_val
        self._pval = int.from_bytes(p_val, "big")
        self._gval = g_val
        self._rem_pub_key = int.from_bytes(rem_pub_key, "big")

    def get_transferable_key(self) -> bytes:
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            the key to be sent to the peer.
        """

        if self._pub_key is None:
            self._create_key_pair()
            assert self._pub_key

        y_val = self._pub_key.public_numbers().y
        return y_val.to_bytes(int(self._pub_key.key_size / 8), "big")

    def get_key_share(self) -> bytes:
        """TLS1.3 alias for get_transferable_key.
        """

        return self.get_transferable_key()

    def set_params(
        self, group_name: Optional[tls.SupportedGroups] = None, **kwargs: Any,
    ) -> None:
        """Sets the group name
        """

        if group_name:
            dh_nbrs = dh_numbers.dh_numbers.get(group_name)
            if dh_nbrs is None:
                raise ValueError(f"No numbers defined for DH-group {group_name.name}")

            self._gval = dh_nbrs.g_val
            self._pval = int.from_bytes(dh_nbrs.p_val, "big")


class EcdhKeyExchange(KeyExchange):
    """Implement an ECDHE key exchange
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._algo: Optional[Type[ec.EllipticCurve]] = kwargs.get("algo")
        self._priv_key: Optional[ec.EllipticCurvePrivateKey] = None
        self._pub_key: Optional[bytes] = None
        self._rem_pub_key: Optional[bytes] = None
        super().__init__(*args, **kwargs)

    def _create_key_pair(self) -> None:
        seed = int.from_bytes(os.urandom(10), "big")
        seed = self._recorder.inject(ec_seed=seed)
        assert self._algo
        self._priv_key = ec.derive_private_key(seed, self._algo())
        pub_key = self._priv_key.public_key()
        self._pub_key = pub_key.public_bytes(
            Encoding.X962, PublicFormat.UncompressedPoint
        )

    def get_shared_secret(self) -> bytes:
        """Get the shared key, i.e., the premaster secret.

        Returns:
            the premaster secret
        """

        if self._priv_key is None:
            self._create_key_pair()

        assert self._algo and self._rem_pub_key and self._priv_key
        rem_pub_key = ec.EllipticCurvePublicKey.from_encoded_point(
            self._algo(), bytes(self._rem_pub_key)
        )
        return self._priv_key.exchange(ec.ECDH(), rem_pub_key)

    def set_remote_key(self, rem_pub_key: bytes, **kwargs: Any) -> None:
        """Sets the remote public key

        Arguments:
            rem_pub_key: the raw remote public key
            **kwargs: unused
        """

        self._rem_pub_key = rem_pub_key

    def get_transferable_key(self) -> bytes:
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            the key to be sent to the peer.
        """

        if self._pub_key is None:
            self._create_key_pair()
            assert self._pub_key

        return self._pub_key

    def get_key_share(self) -> bytes:
        """TLS1.3 alias for get_transferable_key.
        """

        if self._pub_key is None:
            self._create_key_pair()
            assert self._pub_key

        return self._pub_key

    def set_params(
        self,
        group_name: Optional[tls.SupportedGroups] = None,
        algo: Optional[Type[ec.EllipticCurve]] = None,
        **kwargs: Any,
    ) -> None:
        """Sets the versions and the remote public key
        """

        if group_name:
            self._group_name = group_name

        if algo:
            self._algo = algo


class EcdhKeyExchangeCertificate(KeyExchange):
    """Implement the key exchange for ECDHE.
    """

    def set_remote_key(self, rem_pub_key: bytes, **kwargs: Any) -> None:
        """Sets the remote public key

        Arguments:
            rem_pub_key: the raw remote public key
            **kwargs: unused
        """

        pass

    def get_transferable_key(self) -> bytes:
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            the key to be sent to the peer.
        """

        return self._pub_key

    def get_shared_secret(self) -> bytes:
        """Get the shared key, i.e., the premaster secret.

        Returns:
            the premaster secret
        """

        return self._shared_secret

    def set_params(self, rem_public_key: Optional[Any] = None, **kwargs: Any,) -> None:
        """Sets the remote public key
        """

        if rem_public_key:
            seed = self._recorder.inject(ec_seed=int.from_bytes(os.urandom(10), "big"))
            priv_key = ec.derive_private_key(seed, rem_public_key.curve)
            pub_key = priv_key.public_key()
            self._pub_key: bytes = pub_key.public_bytes(
                Encoding.X962, PublicFormat.UncompressedPoint
            )
            self._shared_secret: bytes = priv_key.exchange(ec.ECDH(), rem_public_key)


class XKeyExchange(KeyExchange):
    """Implement the key exchange for X25519 and X448.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._priv_key = None
        self._pub_key = None
        self._rem_pub_key: Optional[bytes] = None
        self._private_key_lib: Union[
            Type[x25519.X25519PrivateKey], Type[x448.X448PrivateKey]
        ]
        self._public_key_lib: Union[
            Type[x25519.X25519PublicKey], Type[x448.X448PublicKey]
        ]

    def _create_key_pair(self):
        if self._recorder.is_injecting():
            priv_bytes = self._recorder.inject(private_key=None)
            self._priv_key = self._private_key_lib.from_private_bytes(priv_bytes)

        else:
            self._priv_key = self._private_key_lib.generate()
            if self._recorder.is_recording():
                priv_bytes = self._priv_key.private_bytes(
                    encoding=Encoding.Raw,
                    format=PrivateFormat.Raw,
                    encryption_algorithm=NoEncryption(),
                )
                self._recorder.trace(private_key=priv_bytes)

        pub_key = self._priv_key.public_key()
        self._pub_key = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)

    def get_shared_secret(self) -> bytes:
        """Get the shared key, i.e., the premaster secret.

        Returns:
            the premaster secret
        """

        if self._priv_key is None:
            self._create_key_pair()
            assert self._priv_key

        rem_pub_key = self._public_key_lib.from_public_bytes(bytes(self._rem_pub_key))
        return self._priv_key.exchange(rem_pub_key)

    def set_remote_key(self, rem_pub_key: bytes, **kwargs: Any) -> None:
        """Sets the remote public key

        Arguments:
            rem_pub_key: the raw remote public key
            **kwargs: unused
        """

        self._rem_pub_key = rem_pub_key

    def get_transferable_key(self) -> bytes:
        """Get the raw bytes of the key to be sent to the peer.

        Returns:
            the key to be sent to the peer.
        """

        if self._priv_key is None:
            self._create_key_pair()
            assert self._pub_key

        return self._pub_key

    def get_key_share(self) -> bytes:
        """TLS1.3 alias for get_transferable_key.
        """

        return self.get_transferable_key()

    def set_params(
        self, group_name: Optional[tls.SupportedGroups] = None, **kwargs: Any,
    ) -> None:
        """Sets the group name
        """

        if group_name is tls.SupportedGroups.X25519:
            self._private_key_lib = x25519.X25519PrivateKey
            self._public_key_lib = x25519.X25519PublicKey

        elif group_name is tls.SupportedGroups.X448:
            self._private_key_lib = x448.X448PrivateKey
            self._public_key_lib = x448.X448PublicKey

        elif group_name:
            raise ValueError(f"the group name {group_name.name} is not expected")


def instantiate_named_group(
    recorder: rec.Recorder, group_name: tls.SupportedGroups
) -> KeyExchange:
    """Create a KeyExchange object according to the given group name.

    Arguments:
        recorder: the recorder object
        group_name: the group name

    Returns:
        the created object
    """

    group = _supported_groups.get(group_name)
    if group is None:
        raise tls.CurveNotSupportedError(
            f"the group {group_name} is not supported", group_name
        )

    key_exchange = group.cls(recorder)
    key_exchange.set_params(group_name=group_name)
    if group.algo:
        key_exchange.set_params(algo=group.algo)

    return key_exchange


class _Group(NamedTuple):
    """Structure for a group
    """

    cls: type
    algo: Optional[type]


_supported_groups = {
    tls.SupportedGroups.SECT163K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT163K1),
    tls.SupportedGroups.SECT163R2: _Group(cls=EcdhKeyExchange, algo=ec.SECT163R2),
    tls.SupportedGroups.SECT233K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT233K1),
    tls.SupportedGroups.SECT233R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT233R1),
    tls.SupportedGroups.SECT283K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT283K1),
    tls.SupportedGroups.SECT283R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT283R1),
    tls.SupportedGroups.SECT409K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT409K1),
    tls.SupportedGroups.SECT409R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT409R1),
    tls.SupportedGroups.SECT571K1: _Group(cls=EcdhKeyExchange, algo=ec.SECT571K1),
    tls.SupportedGroups.SECT571R1: _Group(cls=EcdhKeyExchange, algo=ec.SECT571R1),
    tls.SupportedGroups.SECP192R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP192R1),
    tls.SupportedGroups.SECP224R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP224R1),
    tls.SupportedGroups.SECP256K1: _Group(cls=EcdhKeyExchange, algo=ec.SECP256K1),
    tls.SupportedGroups.SECP256R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP256R1),
    tls.SupportedGroups.SECP384R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP384R1),
    tls.SupportedGroups.SECP521R1: _Group(cls=EcdhKeyExchange, algo=ec.SECP521R1),
    tls.SupportedGroups.BRAINPOOLP256R1: _Group(
        cls=EcdhKeyExchange, algo=ec.BrainpoolP256R1
    ),
    tls.SupportedGroups.BRAINPOOLP384R1: _Group(
        cls=EcdhKeyExchange, algo=ec.BrainpoolP384R1
    ),
    tls.SupportedGroups.BRAINPOOLP512R1: _Group(
        cls=EcdhKeyExchange, algo=ec.BrainpoolP512R1
    ),
    tls.SupportedGroups.X25519: _Group(cls=XKeyExchange, algo=None),
    tls.SupportedGroups.X448: _Group(cls=XKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE2048: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE3072: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE4096: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE6144: _Group(cls=DhKeyExchange, algo=None),
    tls.SupportedGroups.FFDHE8192: _Group(cls=DhKeyExchange, algo=None),
}
